#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Examples to show how to create async EventHubProducerClient/EventHubConsumerClient.
"""
import asyncio
import os
from azure.eventhub import TransportType
from azure.eventhub.aio import EventHubProducerClient, EventHubConsumerClient, EventHubSharedKeyCredential


CONNECTION_STRING = os.environ['EVENT_HUB_CONN_STR']
FULLY_QUALIFIED_NAMESPACE = os.environ['EVENT_HUB_HOSTNAME']
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']
SAS_POLICY = os.environ['EVENT_HUB_SAS_POLICY']
SAS_KEY = os.environ['EVENT_HUB_SAS_KEY']
CONSUMER_GROUP = "$Default"


async def create_producer_client():
    print('Examples showing how to create producer client.')

    # Create producer client from connection string.

    producer_client = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STRING  # connection string contains EventHub name.
    )

    # Illustration of commonly used parameters.
    producer_client = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STRING,
        eventhub_name=EVENTHUB_NAME,  # EventHub name should be specified if it doesn't show up in connection string.
        logging_enable=False,  # To enable network tracing log, set logging_enable to True.
        retry_total=3,  # Retry up to 3 times to re-do failed operations.
        transport_type=TransportType.Amqp  # Use Amqp as the underlying transport protocol.
    )

    # Create producer client from constructor.

    producer_client = EventHubProducerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        credential=EventHubSharedKeyCredential(
            policy=SAS_POLICY,
            key=SAS_KEY
        ),
        logging_enable=False,  # To enable network tracing log, set logging_enable to True.
        retry_total=3,  # Retry up to 3 times to re-do failed operations.
        transport_type=TransportType.Amqp  # Use Amqp as the underlying transport protocol.
    )

    async with producer_client:
        print("Calling producer client get eventhub properties:", await producer_client.get_eventhub_properties())


async def create_consumer_client():
    print('Examples showing how to create consumer client.')

    # Create consumer client from connection string.

    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STRING,  # connection string contains EventHub name.
        consumer_group=CONSUMER_GROUP
    )

    # Illustration of commonly used parameters.
    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STRING,
        consumer_group=CONSUMER_GROUP,
        eventhub_name=EVENTHUB_NAME,  # EventHub name should be specified if it doesn't show up in connection string.
        logging_enable=False,  # To enable network tracing log, set logging_enable to True.
        retry_total=3,  # Retry up to 3 times to re-do failed operations.
        transport_type=TransportType.Amqp  # Use Amqp as the underlying transport protocol.
    )

    # Create consumer client from constructor.

    consumer_client = EventHubConsumerClient(
        fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
        eventhub_name=EVENTHUB_NAME,
        consumer_group=CONSUMER_GROUP,
        credential=EventHubSharedKeyCredential(
            policy=SAS_POLICY,
            key=SAS_KEY
        ),
        logging_enable=False,  # To enable network tracing log, set logging_enable to True.
        retry_total=3,  # Retry up to 3 times to re-do failed operations.
        transport_type=TransportType.Amqp  # Use Amqp as the underlying transport protocol.
    )

    async with consumer_client:
        print("Calling consumer client get eventhub properties:", await consumer_client.get_eventhub_properties())


asyncio.run(create_producer_client())
asyncio.run(create_consumer_client())


# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import uuid
import logging
import time
from urllib.parse import urlparse
import socket
from ssl import SSLError
from typing import Any, Dict, Tuple, Optional, NamedTuple, Union, cast

from ._transport import Transport
from .sasl import SASLTransport, SASLWithWebSocket
from .session import Session
from .performatives import OpenFrame, CloseFrame
from .constants import (
    PORT,
    SECURE_PORT,
    SOCKET_TIMEOUT,
    WEBSOCKET_PORT,
    MAX_CHANNELS,
    MAX_FRAME_SIZE_BYTES,
    HEADER_FRAME,
    WS_TIMEOUT_INTERVAL,
    ConnectionState,
    EMPTY_FRAME,
    TransportType,
)

from .error import ErrorCondition, AMQPConnectionError, AMQPError

_LOGGER = logging.getLogger(__name__)
_CLOSING_STATES = (
    ConnectionState.OC_PIPE,
    ConnectionState.CLOSE_PIPE,
    ConnectionState.DISCARDING,
    ConnectionState.CLOSE_SENT,
    ConnectionState.END,
)


def get_local_timeout(now, idle_timeout, last_frame_received_time):
    # type: (float, float, float) -> bool
    """Check whether the local timeout has been reached since a new incoming frame was received.

    :param float now: The current time to check against.
    :param float idle_timeout: The idle timeout value.
    :param float last_frame_received_time: The time the last frame was received.
    :returns: Whether to shutdown the connection due to timeout.
    :rtype: bool
    """
    if idle_timeout and last_frame_received_time:
        time_since_last_received = now - last_frame_received_time
        return time_since_last_received > idle_timeout
    return False


class Connection(object):  # pylint:disable=too-many-instance-attributes
    """An AMQP Connection.

    :ivar str state: The connection state.
    :param str endpoint: The endpoint to connect to. Must be fully qualified with scheme and port number.
    :keyword str container_id: The ID of the source container. If not set a GUID will be generated.
    :keyword int max_frame_size: Proposed maximum frame size in bytes. Default value is 64kb.
    :keyword int channel_max: The maximum channel number that may be used on the Connection. Default value is 65535.
    :keyword int idle_timeout: Connection idle time-out in seconds.
    :keyword list(str) outgoing_locales: Locales available for outgoing text.
    :keyword list(str) incoming_locales: Desired locales for incoming text in decreasing level of preference.
    :keyword list(str) offered_capabilities: The extension capabilities the sender supports.
    :keyword list(str) desired_capabilities: The extension capabilities the sender may use if the receiver supports
    :keyword dict properties: Connection properties.
    :keyword bool allow_pipelined_open: Allow frames to be sent on the connection before a response Open frame
     has been received. Default value is `True`.
    :keyword float idle_timeout_empty_frame_send_ratio: Portion of the idle timeout time to wait before sending an
     empty frame. The default portion is 50% of the idle timeout value (i.e. `0.5`).
    :keyword float idle_wait_time: The time in seconds to sleep while waiting for a response from the endpoint.
     Default value is `0.1`.
    :keyword bool network_trace: Whether to log the network traffic. Default value is `False`. If enabled, frames
     will be logged at the logging.INFO level.
    :keyword str transport_type: Determines if the transport type is Amqp or AmqpOverWebSocket.
     Defaults to TransportType.Amqp. It will be AmqpOverWebSocket if using http_proxy.
    :keyword Dict http_proxy: HTTP proxy settings. This must be a dictionary with the following
     keys: `'proxy_hostname'` (str value) and `'proxy_port'` (int value). When using these settings,
     the transport_type would be AmqpOverWebSocket.
     Additionally the following keys may also be present: `'username', 'password'`.
    :keyword float socket_timeout: The maximum time in seconds that the underlying socket in the transport should
     wait when reading or writing data before timing out. The default value is 0.2 (for transport type Amqp),
     and 1 for transport type AmqpOverWebsocket.
    """

    def __init__(self, endpoint, **kwargs):  # pylint:disable=too-many-statements
        # type(str, Any) -> None
        parsed_url = urlparse(endpoint)
        self._hostname = parsed_url.hostname
        endpoint = self._hostname
        if parsed_url.port:
            self._port = parsed_url.port
        elif parsed_url.scheme == "amqps":
            self._port = SECURE_PORT
        else:
            self._port = PORT
        self.state = None  # type: Optional[ConnectionState]

        # Custom Endpoint
        custom_endpoint_address = kwargs.get("custom_endpoint_address")
        custom_endpoint = None
        if custom_endpoint_address:
            custom_parsed_url = urlparse(custom_endpoint_address)
            custom_port = custom_parsed_url.port or WEBSOCKET_PORT
            custom_endpoint = f"{custom_parsed_url.hostname}:{custom_port}{custom_parsed_url.path}"
        self._container_id = kwargs.pop("container_id", None) or str(uuid.uuid4())  # type: str
        self._network_trace = kwargs.get("network_trace", False)
        self._network_trace_params = {"amqpConnection": self._container_id, "amqpSession": None, "amqpLink": None}

        transport = kwargs.get("transport")
        self._transport_type = kwargs.pop("transport_type", TransportType.Amqp)
        self._socket_timeout = kwargs.pop("socket_timeout", None)

        if self._transport_type.value == TransportType.Amqp.value and self._socket_timeout is None:
            self._socket_timeout = SOCKET_TIMEOUT
        elif (self._transport_type.value == TransportType.AmqpOverWebsocket.value and
              self._socket_timeout is None):
            self._socket_timeout = WS_TIMEOUT_INTERVAL

        if transport:
            self._transport = transport
        elif "sasl_credential" in kwargs:
            sasl_transport = SASLTransport
            if self._transport_type.name == "AmqpOverWebsocket" or kwargs.get("http_proxy"):
                sasl_transport = SASLWithWebSocket
                endpoint = parsed_url.hostname + parsed_url.path
            self._transport = sasl_transport(
                host=endpoint,
                credential=kwargs["sasl_credential"],
                custom_endpoint=custom_endpoint,
                socket_timeout=self._socket_timeout,
                network_trace_params=self._network_trace_params,
                **kwargs
            )
        else:
            self._transport = Transport(
                parsed_url.netloc,
                transport_type=self._transport_type,
                socket_timeout=self._socket_timeout,
                network_trace_params=self._network_trace_params,
                **kwargs)
        self._max_frame_size = kwargs.pop("max_frame_size", MAX_FRAME_SIZE_BYTES)  # type: int
        self._remote_max_frame_size = None  # type: Optional[int]
        self._channel_max = kwargs.pop("channel_max", MAX_CHANNELS)  # type: int
        self._idle_timeout = kwargs.pop("idle_timeout", None)  # type: Optional[int]
        self._outgoing_locales = kwargs.pop("outgoing_locales", None)  # type: Optional[List[str]]
        self._incoming_locales = kwargs.pop("incoming_locales", None)  # type: Optional[List[str]]
        self._offered_capabilities = None  # type: Optional[str]
        self._desired_capabilities = kwargs.pop("desired_capabilities", None)  # type: Optional[str]
        self._properties = kwargs.pop("properties", None)  # type: Optional[Dict[str, str]]
        self._remote_properties: Optional[Dict[str, str]] = None

        self._allow_pipelined_open = kwargs.pop("allow_pipelined_open", True)  # type: bool
        self._remote_idle_timeout = None  # type: Optional[int]
        self._remote_idle_timeout_send_frame = None  # type: Optional[int]
        self._idle_timeout_empty_frame_send_ratio = kwargs.get("idle_timeout_empty_frame_send_ratio", 0.5)
        self._last_frame_received_time = None  # type: Optional[float]
        self._last_frame_sent_time = None  # type: Optional[float]
        self._idle_wait_time = kwargs.get("idle_wait_time", 0.1)  # type: float
        self._error = None
        self._outgoing_endpoints = {}  # type: Dict[int, Session]
        self._incoming_endpoints = {}  # type: Dict[int, Session]

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *args):
        self.close()

    def _set_state(self, new_state):
        # type: (ConnectionState) -> None
        """Update the connection state.
        :param ~pyamqp.constants.ConnectionState new_state: The new state of the connection.
        """
        if new_state is None:
            return
        previous_state = self.state
        self.state = new_state
        _LOGGER.info(
            "Connection state changed: %r -> %r",
            previous_state,
            new_state,
            extra=self._network_trace_params
        )
        for session in self._outgoing_endpoints.values():
            session._on_connection_state_change()  # pylint:disable=protected-access

    def _connect(self):
        # type: () -> None
        """Initiate the connection.

        If `allow_pipelined_open` is enabled, the incoming response header will be processed immediately
        and the state on exiting will be HDR_EXCH. Otherwise, the function will return before waiting for
        the response header and the final state will be HDR_SENT.

        :raises ValueError: If a reciprocating protocol header is not received during negotiation.
        """
        try:
            if not self.state:
                self._transport.connect()
                self._set_state(ConnectionState.START)
            self._transport.negotiate()
            self._outgoing_header()
            self._set_state(ConnectionState.HDR_SENT)
            if not self._allow_pipelined_open:
                # TODO: List/tuple expected as variable args
                self._read_frame(wait=True)
                if self.state != ConnectionState.HDR_EXCH:
                    self._disconnect()
                    raise ValueError("Did not receive reciprocal protocol header. Disconnecting.")
            else:
                self._set_state(ConnectionState.HDR_SENT)
        except (OSError, IOError, SSLError, socket.error) as exc:
            # FileNotFoundError is being raised for exception parity with uamqp when invalid
            # `connection_verify` file path is passed in. Remove later when resolving issue #27128.
            if isinstance(exc, FileNotFoundError) and exc.filename and "ca_certs" in exc.filename:
                raise
            raise AMQPConnectionError(
                ErrorCondition.SocketError,
                description="Failed to initiate the connection due to exception: " + str(exc),
                error=exc,
            ) from exc

    def _disconnect(self):
        # type: () -> None
        """Disconnect the transport and set state to END."""
        if self.state == ConnectionState.END:
            return
        self._set_state(ConnectionState.END)
        self._transport.close()

    def _can_read(self):
        # type: () -> bool
        """Whether the connection is in a state where it is legal to read for incoming frames.
        :return: True if the connection is in a state where it is legal to read for incoming frames.
        :rtype: bool
        """
        return self.state not in (ConnectionState.CLOSE_RCVD, ConnectionState.END)

    def _read_frame(
        self, wait: Union[bool, float] = True, **kwargs: Any
    ) -> bool:
        """Read an incoming frame from the transport.

        :param Union[bool, float] wait: Whether to block on the socket while waiting for an incoming frame.
         The default value is `False`, where the frame will block for the configured timeout only (0.1 seconds).
         If set to `True`, socket will block indefinitely. If set to a timeout value in seconds, the socket will
         block for at most that value.
        :rtype: Tuple[int, Optional[Tuple[int, NamedTuple]]]
        :returns: A tuple with the incoming channel number, and the frame in the form or a tuple of performative
         descriptor and field values.
        """
        # Since we use `sock.settimeout()` in the transport for reading/writing, that acts as a
        # "block with timeout" when we pass in a timeout value. If `wait` is float value, then
        # timeout was set during socket init.
        if wait is not True:    # wait is float/int/False
            new_frame = self._transport.receive_frame(**kwargs)
        else:
            with self._transport.block():
                new_frame = self._transport.receive_frame(**kwargs)
        return self._process_incoming_frame(*new_frame)

    def _can_write(self):
        # type: () -> bool
        """Whether the connection is in a state where it is legal to write outgoing frames.
        :return: Whether the connection is in a state where it is legal to write outgoing frames.
        :rtype: bool
        """
        return self.state not in _CLOSING_STATES

    def _send_frame(self, channel, frame, **kwargs):
        # type: (int, NamedTuple, Any) -> None
        """Send a frame over the connection.

        :param int channel: The outgoing channel number.
        :param NamedTuple frame: The outgoing frame.
        :rtype: None
        """
        try:
            raise self._error
        except TypeError:
            pass

        if self._can_write():
            try:
                self._last_frame_sent_time = time.time()
                # Since we use `sock.settimeout()` in the transport for reading/writing,
                # that acts as a "block with timeout" when we pass in a timeout value.
                self._transport.send_frame(channel, frame, **kwargs)
            except (OSError, IOError, SSLError, socket.error) as exc:
                self._error = AMQPConnectionError(
                    ErrorCondition.SocketError,
                    description="Can not send frame out due to exception: " + str(exc),
                    error=exc,
                )
            except Exception:  # pylint:disable=try-except-raise
                raise
        else:
            _LOGGER.info("Cannot write frame in current state: %r", self.state, extra=self._network_trace_params)

    def _get_next_outgoing_channel(self):
        # type: () -> int
        """Get the next available outgoing channel number within the max channel limit.

        :raises ValueError: If maximum channels has been reached.
        :returns: The next available outgoing channel number.
        :rtype: int
        """
        if (len(self._incoming_endpoints) + len(self._outgoing_endpoints)) >= self._channel_max:
            raise ValueError("Maximum number of channels ({}) has been reached.".format(self._channel_max))
        next_channel = next(i for i in range(1, self._channel_max) if i not in self._outgoing_endpoints)
        return next_channel

    def _outgoing_empty(self):
        # type: () -> None
        """Send an empty frame to prevent the connection from reaching an idle timeout."""
        if self._network_trace:
            _LOGGER.debug("-> EmptyFrame()", extra=self._network_trace_params)
        try:
            raise self._error
        except TypeError:
            pass
        try:
            if self._can_write():
                self._transport.write(EMPTY_FRAME)
                self._last_frame_sent_time = time.time()
        except (OSError, IOError, SSLError, socket.error) as exc:
            self._error = AMQPConnectionError(
                ErrorCondition.SocketError,
                description="Can not send empty frame due to exception: " + str(exc),
                error=exc,
            )
        except Exception:  # pylint:disable=try-except-raise
            raise

    def _outgoing_header(self):
        # type: () -> None
        """Send the AMQP protocol header to initiate the connection."""
        self._last_frame_sent_time = time.time()
        if self._network_trace:
            _LOGGER.debug("-> Header(%r)", HEADER_FRAME, extra=self._network_trace_params)
        self._transport.write(HEADER_FRAME)

    def _incoming_header(self, _, frame):
        # type: (int, bytes) -> None
        """Process an incoming AMQP protocol header and update the connection state.

        :param int _: Ignored.
        :param bytes frame: The incoming frame.
        """
        if self._network_trace:
            _LOGGER.debug("<- Header(%r)", frame, extra=self._network_trace_params)
        if self.state == ConnectionState.START:
            self._set_state(ConnectionState.HDR_RCVD)
        elif self.state == ConnectionState.HDR_SENT:
            self._set_state(ConnectionState.HDR_EXCH)
        elif self.state == ConnectionState.OPEN_PIPE:
            self._set_state(ConnectionState.OPEN_SENT)

    def _outgoing_open(self):
        # type: () -> None
        """Send an Open frame to negotiate the AMQP connection functionality."""
        open_frame = OpenFrame(
            container_id=self._container_id,
            hostname=self._hostname,
            max_frame_size=self._max_frame_size,
            channel_max=self._channel_max,
            idle_timeout=self._idle_timeout * 1000 if self._idle_timeout else None,  # Convert to milliseconds
            outgoing_locales=self._outgoing_locales,
            incoming_locales=self._incoming_locales,
            offered_capabilities=self._offered_capabilities if self.state == ConnectionState.OPEN_RCVD else None,
            desired_capabilities=self._desired_capabilities if self.state == ConnectionState.HDR_EXCH else None,
            properties=self._properties,
        )
        if self._network_trace:
            _LOGGER.debug("-> %r", open_frame, extra=self._network_trace_params)
        self._send_frame(0, open_frame)

    def _incoming_open(self, channel, frame):
        # type: (int, Tuple[Any, ...]) -> None
        """Process incoming Open frame to finish the connection negotiation.

        The incoming frame format is::

            - frame[0]: container_id (str)
            - frame[1]: hostname (str)
            - frame[2]: max_frame_size (int)
            - frame[3]: channel_max (int)
            - frame[4]: idle_timeout (Optional[int])
            - frame[5]: outgoing_locales (Optional[List[bytes]])
            - frame[6]: incoming_locales (Optional[List[bytes]])
            - frame[7]: offered_capabilities (Optional[List[bytes]])
            - frame[8]: desired_capabilities (Optional[List[bytes]])
            - frame[9]: properties (Optional[Dict[bytes, bytes]])

        :param int channel: The incoming channel number.
        :param frame: The incoming Open frame.
        :type frame: Tuple[Any, ...]
        :rtype: None
        """
        # TODO: Add type hints for full frame tuple contents.
        if self._network_trace:
            _LOGGER.debug("<- %r", OpenFrame(*frame), extra=self._network_trace_params)
        if channel != 0:
            _LOGGER.error("OPEN frame received on a channel that is not 0.", extra=self._network_trace_params)
            self.close(
                error=AMQPError(
                    condition=ErrorCondition.NotAllowed, description="OPEN frame received on a channel that is not 0."
                )
            )
            self._set_state(ConnectionState.END)
        if self.state == ConnectionState.OPENED:
            _LOGGER.error("OPEN frame received in the OPENED state.", extra=self._network_trace_params)
            self.close()
        if frame[4]:
            self._remote_idle_timeout = frame[4] / 1000  # Convert to seconds
            self._remote_idle_timeout_send_frame = (
                self._idle_timeout_empty_frame_send_ratio * self._remote_idle_timeout
            )

        if frame[2] < 512:
            # Max frame size is less than supported minimum.
            # If any of the values in the received open frame are invalid then the connection shall be closed.
            # The error amqp:invalid-field shall be set in the error.condition field of the CLOSE frame.
            self.close(
                error=AMQPError(
                    condition=ErrorCondition.InvalidField,
                    description="Failed parsing OPEN frame: Max frame size is less than supported minimum.",
                )
            )
            _LOGGER.error(
                "Failed parsing OPEN frame: Max frame size is less than supported minimum.",
                extra=self._network_trace_params
            )
            return
        self._remote_max_frame_size = frame[2]
        self._remote_properties = frame[9]
        if self.state == ConnectionState.OPEN_SENT:
            self._set_state(ConnectionState.OPENED)
        elif self.state == ConnectionState.HDR_EXCH:
            self._set_state(ConnectionState.OPEN_RCVD)
            self._outgoing_open()
            self._set_state(ConnectionState.OPENED)
        else:
            self.close(
                error=AMQPError(
                    condition=ErrorCondition.IllegalState,
                    description=f"connection is an illegal state: {self.state}",
                )
            )
            _LOGGER.error("Connection is an illegal state: %r", self.state, extra=self._network_trace_params)

    def _outgoing_close(self, error=None):
        # type: (Optional[AMQPError]) -> None
        """Send a Close frame to shutdown connection with optional error information.
        :param ~pyamqp.error.AMQPError or None error: Optional error information.
        """
        close_frame = CloseFrame(error=error)
        if self._network_trace:
            _LOGGER.debug("-> %r", close_frame, extra=self._network_trace_params)
        self._send_frame(0, close_frame)

    def _incoming_close(self, channel, frame):
        # type: (int, Tuple[Any, ...]) -> None
        """Process incoming Open frame to finish the connection negotiation.

        The incoming frame format is::

            - frame[0]: error (Optional[AMQPError])

        :param int channel: The incoming channel number.
        :param tuple frame: The incoming Close frame.
        """
        if self._network_trace:
            _LOGGER.debug("<- %r", CloseFrame(*frame), extra=self._network_trace_params)
        disconnect_states = [
            ConnectionState.HDR_RCVD,
            ConnectionState.HDR_EXCH,
            ConnectionState.OPEN_RCVD,
            ConnectionState.CLOSE_SENT,
            ConnectionState.DISCARDING,
        ]
        if self.state in disconnect_states:
            self._disconnect()
            return

        close_error = None
        if channel > self._channel_max:
            _LOGGER.error(
                "CLOSE frame received on a channel greated than support max.",
                extra=self._network_trace_params
            )
            close_error = AMQPError(condition=ErrorCondition.InvalidField, description="Invalid channel", info=None)

        self._set_state(ConnectionState.CLOSE_RCVD)
        self._outgoing_close(error=close_error)
        self._disconnect()

        if frame[0]:
            self._error = AMQPConnectionError(
                condition=frame[0][0], description=frame[0][1], info=frame[0][2]
            )
            _LOGGER.error(
                "Connection closed with error: %r", frame[0],
                extra=self._network_trace_params
            )


    def _incoming_begin(self, channel, frame):
        # type: (int, Tuple[Any, ...]) -> None
        """Process incoming Begin frame to finish negotiating a new session.

        The incoming frame format is::

            - frame[0]: remote_channel (int)
            - frame[1]: next_outgoing_id (int)
            - frame[2]: incoming_window (int)
            - frame[3]: outgoing_window (int)
            - frame[4]: handle_max (int)
            - frame[5]: offered_capabilities (Optional[List[bytes]])
            - frame[6]: desired_capabilities (Optional[List[bytes]])
            - frame[7]: properties (Optional[Dict[bytes, bytes]])

        :param int channel: The incoming channel number.
        :param frame: The incoming Begin frame.
        :type frame: Tuple[Any, ...]
        :rtype: None
        """
        try:
            existing_session = self._outgoing_endpoints[frame[0]]
            self._incoming_endpoints[channel] = existing_session
            self._incoming_endpoints[channel]._incoming_begin(  # pylint:disable=protected-access
                frame
            )
        except KeyError:
            new_session = Session.from_incoming_frame(self, channel)
            self._incoming_endpoints[channel] = new_session
            new_session._incoming_begin(frame)  # pylint:disable=protected-access

    def _incoming_end(self, channel, frame):
        # type: (int, Tuple[Any, ...]) -> None
        """Process incoming End frame to close a session.

        The incoming frame format is::

            - frame[0]: error (Optional[AMQPError])

        :param int channel: The incoming channel number.
        :param frame: The incoming End frame.
        :type frame: Tuple[Any, ...]
        :rtype: None
        """
        try:
            self._incoming_endpoints[channel]._incoming_end(frame)  # pylint:disable=protected-access
            outgoing_channel = self._incoming_endpoints[channel].channel
            self._incoming_endpoints.pop(channel)
            self._outgoing_endpoints.pop(outgoing_channel)
        except KeyError:
            #close the connection
            self.close(
                error=AMQPError(
                    condition=ErrorCondition.ConnectionCloseForced,
                    description="Invalid channel number received"
                ))
            _LOGGER.error(
                "END frame received on invalid channel. Closing connection.",
                extra=self._network_trace_params
            )
            return

    def _process_incoming_frame(self, channel, frame):  # pylint:disable=too-many-return-statements
        # type: (int, Optional[Union[bytes, Tuple[int, Tuple[Any, ...]]]]) -> bool
        """Process an incoming frame, either directly or by passing to the necessary Session.

        :param int channel: The channel the frame arrived on.
        :param frame: A tuple containing the performative descriptor and the field values of the frame.
         This parameter can be None in the case of an empty frame or a socket timeout.
        :type frame: Optional[Tuple[int, NamedTuple]]
        :rtype: bool
        :returns: A boolean to indicate whether more frames in a batch can be processed or whether the
         incoming frame has altered the state. If `True` is returned, the state has changed and the batch
         should be interrupted.
        """
        try:
            performative, fields = cast(Union[bytes, Tuple], frame)
        except TypeError:
            return True  # Empty Frame or socket timeout
        fields = cast(Tuple[Any, ...], fields)
        try:
            self._last_frame_received_time = time.time()
            if performative == 20:
                self._incoming_endpoints[channel]._incoming_transfer(  # pylint:disable=protected-access
                    fields
                )
                return False
            if performative == 21:
                self._incoming_endpoints[channel]._incoming_disposition(  # pylint:disable=protected-access
                    fields
                )
                return False
            if performative == 19:
                self._incoming_endpoints[channel]._incoming_flow(  # pylint:disable=protected-access
                    fields
                )
                return False
            if performative == 18:
                self._incoming_endpoints[channel]._incoming_attach(  # pylint:disable=protected-access
                    fields
                )
                return False
            if performative == 22:
                self._incoming_endpoints[channel]._incoming_detach(  # pylint:disable=protected-access
                    fields
                )
                return True
            if performative == 17:
                self._incoming_begin(channel, fields)
                return True
            if performative == 23:
                self._incoming_end(channel, fields)
                return True
            if performative == 16:
                self._incoming_open(channel, fields)
                return True
            if performative == 24:
                self._incoming_close(channel, fields)
                return True
            if performative == 0:
                self._incoming_header(channel, cast(bytes, fields))
                return True
            if performative == 1:
                return False
            _LOGGER.error("Unrecognized incoming frame: %r", frame, extra=self._network_trace_params)
            return True
        except KeyError:
            return True  # TODO: channel error

    def _process_outgoing_frame(self, channel, frame):
        # type: (int, NamedTuple) -> None
        """Send an outgoing frame if the connection is in a legal state.

        :param int channel: The channel to send the frame on.
        :param NamedTuple frame: The frame to send.
        :return: None
        :rtype: None
        :raises ValueError: If the connection is not open or not in a valid state.
        """
        if not self._allow_pipelined_open and self.state in [
            ConnectionState.OPEN_PIPE,
            ConnectionState.OPEN_SENT,
        ]:
            raise ValueError("Connection not configured to allow pipeline send.")
        if self.state not in [
            ConnectionState.OPEN_PIPE,
            ConnectionState.OPEN_SENT,
            ConnectionState.OPENED,
        ]:
            raise ValueError("Connection not open.")
        now = time.time()
        if get_local_timeout(
            now,
            cast(float, self._idle_timeout),
            cast(float, self._last_frame_received_time),
        ) or self._get_remote_timeout(now):
            _LOGGER.info(
                "No frame received for the idle timeout. Closing connection.",
                extra=self._network_trace_params
            )
            self.close(
                error=AMQPError(
                    condition=ErrorCondition.ConnectionCloseForced,
                    description="No frame received for the idle timeout.",
                ),
                wait=False,
            )
            return
        self._send_frame(channel, frame)

    def _get_remote_timeout(self, now):
        # type: (float) -> bool
        """Check whether the local connection has reached the remote endpoints idle timeout since
        the last outgoing frame was sent.

        If the time since the last since frame is greater than the allowed idle interval, an Empty
        frame will be sent to maintain the connection.

        :param float now: The current time to check against.
        :rtype: bool
        :returns: Whether the local connection should be shutdown due to timeout.
        """
        if self._remote_idle_timeout and self._last_frame_sent_time:
            time_since_last_sent = now - self._last_frame_sent_time
            if time_since_last_sent > cast(int, self._remote_idle_timeout_send_frame):
                self._outgoing_empty()
        return False

    def _wait_for_response(self, wait, end_state):
        # type: (Union[bool, float], ConnectionState) -> None
        """Wait for an incoming frame to be processed that will result in a desired state change.

        :param wait: Whether to wait for an incoming frame to be processed. Can be set to `True` to wait
         indefinitely, or an int to wait for a specified amount of time (in seconds). To not wait, set to `False`.
        :type wait: bool or float
        :param ConnectionState end_state: The desired end state to wait until.
        :rtype: None
        """
        if wait is True:
            self.listen(wait=False)
            while self.state != end_state:
                time.sleep(self._idle_wait_time)
                self.listen(wait=False)
        elif wait:
            self.listen(wait=False)
            timeout = time.time() + wait
            while self.state != end_state:
                if time.time() >= timeout:
                    break
                time.sleep(self._idle_wait_time)
                self.listen(wait=False)

    def listen(self, wait=False, batch=1, **kwargs):
        # type: (Union[float, int, bool], int, Any) -> None
        """Listen on the socket for incoming frames and process them.

        :param wait: Whether to block on the socket until a frame arrives. If set to `True`, socket will
         block indefinitely. Alternatively, if set to a time in seconds, the socket will block for at most
         the specified timeout. Default value is `False`, where the socket will block for its configured read
         timeout (by default 0.2 seconds).
        :type wait: int or float or bool
        :param int batch: The number of frames to attempt to read and process before returning. The default value
         is 1, i.e. process frames one-at-a-time. A higher value should only be used when a receiver is established
         and is processing incoming Transfer frames.
        :rtype: None
        """
        try:
            raise self._error
        except TypeError:
            pass
        try:
            if self.state not in _CLOSING_STATES:
                now = time.time()
                if get_local_timeout(
                    now,
                    cast(float, self._idle_timeout),
                    cast(float, self._last_frame_received_time),
                ) or self._get_remote_timeout(
                    now
                ):
                    _LOGGER.info(
                        "No frame received for the idle timeout. Closing connection.",
                        extra=self._network_trace_params
                    )
                    self.close(
                        error=AMQPError(
                            condition=ErrorCondition.ConnectionCloseForced,
                            description="No frame received for the idle timeout.",
                        ),
                        wait=False,
                    )
                    return
            if self.state == ConnectionState.END:
                self._error = AMQPConnectionError(
                    condition=ErrorCondition.ConnectionCloseForced, description="Connection was already closed."
                )
                return
            for _ in range(batch):
                if self._can_read():
                    if self._read_frame(wait=wait, **kwargs):
                        break
                else:
                    _LOGGER.info(
                        "Connection cannot read frames in this state: %r",
                        self.state,
                        extra=self._network_trace_params
                    )
                    break
        except (OSError, IOError, SSLError, socket.error) as exc:
            self._error = AMQPConnectionError(
                ErrorCondition.SocketError,
                description="Can not read frame due to exception: " + str(exc),
                error=exc,
            )
        except Exception:  # pylint:disable=try-except-raise
            raise

    def create_session(self, **kwargs):
        # type: (Any) -> Session
        """Create a new session within this connection.

        :keyword str name: The name of the connection. If not set a GUID will be generated.
        :keyword int next_outgoing_id: The transfer-id of the first transfer id the sender will send.
         Default value is 0.
        :keyword int incoming_window: The initial incoming-window of the Session. Default value is 1.
        :keyword int outgoing_window: The initial outgoing-window of the Session. Default value is 1.
        :keyword int handle_max: The maximum handle value that may be used on the session. Default value is 4294967295.
        :keyword list(str) offered_capabilities: The extension capabilities the session supports.
        :keyword list(str) desired_capabilities: The extension capabilities the session may use if
         the endpoint supports it.
        :keyword dict properties: Session properties.
        :keyword bool allow_pipelined_open: Allow frames to be sent on the connection before a response Open frame
         has been received. Default value is that configured for the connection.
        :keyword float idle_wait_time: The time in seconds to sleep while waiting for a response from the endpoint.
         Default value is that configured for the connection.
        :keyword bool network_trace: Whether to log the network traffic of this session. If enabled, frames
         will be logged at the logging.INFO level. Default value is that configured for the connection.

        :return: A new Session.
        :rtype: ~pyamqp._session.Session
        """
        assigned_channel = self._get_next_outgoing_channel()
        kwargs["allow_pipelined_open"] = self._allow_pipelined_open
        kwargs["idle_wait_time"] = self._idle_wait_time
        session = Session(
            self,
            assigned_channel,
            network_trace=kwargs.pop("network_trace", self._network_trace),
            network_trace_params=dict(self._network_trace_params),
            **kwargs,
        )
        self._outgoing_endpoints[assigned_channel] = session
        return session

    def open(self, wait=False):
        # type: (bool) -> None
        """Send an Open frame to start the connection.

        Alternatively, this will be called on entering a Connection context manager.

        :param bool wait: Whether to wait to receive an Open response from the endpoint. Default is `False`.
        :raises ValueError: If `wait` is set to `False` and `allow_pipelined_open` is disabled.
        :rtype: None
        """
        self._connect()
        self._outgoing_open()
        if self.state == ConnectionState.HDR_EXCH:
            self._set_state(ConnectionState.OPEN_SENT)
        elif self.state == ConnectionState.HDR_SENT:
            self._set_state(ConnectionState.OPEN_PIPE)
        if wait:
            self._wait_for_response(wait, ConnectionState.OPENED)
        elif not self._allow_pipelined_open:
            raise ValueError(
                "Connection has been configured to not allow piplined-open. Please set 'wait' parameter."
            )


    def close(self, error=None, wait=False):
        # type: (Optional[AMQPError], bool) -> None
        """Close the connection and disconnect the transport.

        Alternatively this method will be called on exiting a Connection context manager.

        :param ~uamqp.AMQPError error: Optional error information to include in the close request.
        :param bool wait: Whether to wait for a service Close response. Default is `False`.
        :rtype: None
        """
        try:
            if self.state in [
                ConnectionState.END,
                ConnectionState.CLOSE_SENT,
                ConnectionState.DISCARDING,
            ]:
                return
            self._outgoing_close(error=error)
            if error:
                self._error = AMQPConnectionError(
                    condition=error.condition,
                    description=error.description,
                    info=error.info,
                )
            if self.state == ConnectionState.OPEN_PIPE:
                self._set_state(ConnectionState.OC_PIPE)
            elif self.state == ConnectionState.OPEN_SENT:
                self._set_state(ConnectionState.CLOSE_PIPE)
            elif error:
                self._set_state(ConnectionState.DISCARDING)
            else:
                self._set_state(ConnectionState.CLOSE_SENT)
            self._wait_for_response(wait, ConnectionState.END)
        except Exception as exc:  # pylint:disable=broad-except
            # If error happened during closing, ignore the error and set state to END
            _LOGGER.info("An error occurred when closing the connection: %r", exc, extra=self._network_trace_params)
            self._set_state(ConnectionState.END)
        finally:
            self._disconnect()



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import pytest

from azure.eventhub._pyamqp import SendClient, ReceiveClient

def test_send_client_creation():

    sender = SendClient(
        "fake.host.com",
        "fake_eh",
        auth="my_fake_auth"
    )
    assert sender.target == "fake_eh"
    assert sender._auth == "my_fake_auth"


def test_receive_client_creation():

    receiver = ReceiveClient(
        "fake.host.com",
        "fake_eh",
        auth="my_fake_auth"
    )
    assert receiver.source == "fake_eh"
    assert receiver._auth == "my_fake_auth"



import pytest
import asyncio

from azure.core.settings import settings
from azure.core.tracing import SpanKind
from azure.eventhub import EventData
from azure.eventhub.aio import EventHubConsumerClient
from azure.eventhub.aio._eventprocessor.in_memory_checkpoint_store import InMemoryCheckpointStore
from azure.eventhub._constants import ALL_PARTITIONS


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_storage_checkpoint_async(connstr_senders, uamqp_transport, checkpoint_store_aio, live_eventhub, resource_mgmt_client):
    connection_str, senders = connstr_senders

    for i in range(10):
        senders[0].send(EventData("Test EventData"))
        senders[1].send(EventData("Test EventData"))

    try:
        await checkpoint_store_aio._container_client.create_container()
    except:
        pass

    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', checkpoint_store=checkpoint_store_aio, uamqp_transport=uamqp_transport)

    sequence_numbers_0 = []
    sequence_numbers_1 = []
    async def on_event(partition_context, event):
        await partition_context.update_checkpoint(event)
        sequence_num = event.sequence_number
        if partition_context.partition_id == "0":
            if sequence_num in sequence_numbers_0:
                assert False
            sequence_numbers_0.append(sequence_num)
        else:
            if sequence_num in sequence_numbers_1:
                assert False
            sequence_numbers_1.append(sequence_num)

    async with client:
        task = asyncio.ensure_future(
            client.receive(on_event, starting_position="-1"))
        # Update the eventhub
        eventhub = resource_mgmt_client.event_hubs.get(
            live_eventhub["resource_group"],
            live_eventhub["namespace"],
            live_eventhub["event_hub"]
        )
        properties = eventhub.as_dict()
        if properties["message_retention_in_days"] == 1:
            properties["message_retention_in_days"] = 2
        else:
            properties["message_retention_in_days"] = 1
        resource_mgmt_client.event_hubs.create_or_update(
            live_eventhub["resource_group"],
            live_eventhub["namespace"],
            live_eventhub["event_hub"],
            properties
        )
        await asyncio.sleep(10)

 
    await task
    assert len(sequence_numbers_0) == 10
    assert len(sequence_numbers_1) == 10

    try:
        await checkpoint_store_aio._container_client.delete_container()
    except:
        pass

@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_no_partition_async(connstr_senders, uamqp_transport):
    connection_str, senders = connstr_senders
    senders[0].send(EventData("Test EventData"))
    senders[1].send(EventData("Test EventData"))
    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)

    async def on_event(partition_context, event):
        on_event.received += 1
        await partition_context.update_checkpoint(event)
        on_event.namespace = partition_context.fully_qualified_namespace
        on_event.eventhub_name = partition_context.eventhub_name
        on_event.consumer_group = partition_context.consumer_group
        on_event.offset = event.offset
        on_event.sequence_number = event.sequence_number

    on_event.received = 0

    on_event.namespace = None
    on_event.eventhub_name = None
    on_event.consumer_group = None
    on_event.offset = None
    on_event.sequence_number = None

    async with client:
        task = asyncio.ensure_future(
            client.receive(on_event, starting_position="-1"))
        await asyncio.sleep(10)
        assert on_event.received == 2

        checkpoints = await list(client._event_processors.values())[0]._checkpoint_store.list_checkpoints(
            on_event.namespace, on_event.eventhub_name, on_event.consumer_group
        )
        assert len([checkpoint for checkpoint in checkpoints if checkpoint["offset"] == on_event.offset]) > 0
        assert len(
            [checkpoint for checkpoint in checkpoints if checkpoint["sequence_number"] == on_event.sequence_number]) > 0

    await task


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_partition_async(connstr_senders, uamqp_transport):
    connection_str, senders = connstr_senders
    senders[0].send(EventData("Test EventData"))
    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)

    async def on_event(partition_context, event):
        assert partition_context.partition_id == "0"
        assert partition_context.consumer_group == "$default"
        assert partition_context.fully_qualified_namespace in connection_str
        assert partition_context.eventhub_name == senders[0]._client.eventhub_name
        on_event.received += 1

    on_event.received = 0
    async with client:
        task = asyncio.ensure_future(
            client.receive(on_event, partition_id="0", starting_position="-1"))
        await asyncio.sleep(10)
        assert on_event.received == 1
    await task


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_load_balancing_async(connstr_senders, uamqp_transport):
    connection_str, senders = connstr_senders
    cs = InMemoryCheckpointStore()
    client1 = EventHubConsumerClient.from_connection_string(
        connection_str, consumer_group='$default', checkpoint_store=cs, load_balancing_interval=1, uamqp_transport=uamqp_transport)
    client2 = EventHubConsumerClient.from_connection_string(
        connection_str, consumer_group='$default', checkpoint_store=cs, load_balancing_interval=1, uamqp_transport=uamqp_transport)

    async def on_event(partition_context, event):
        pass

    async with client1, client2:
        task1 = asyncio.ensure_future(
            client1.receive(on_event, starting_position="-1"))
        await asyncio.sleep(3.3)
        task2 = asyncio.ensure_future(
            client2.receive(on_event, starting_position="-1"))
        await asyncio.sleep(20)
        assert len(client1._event_processors[("$default", ALL_PARTITIONS)]._tasks) == 1
        assert len(client2._event_processors[("$default", ALL_PARTITIONS)]._tasks) == 1
    await task1
    await task2


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_batch_no_max_wait_time_async(connstr_senders, uamqp_transport):
    '''Test whether callback is called when max_wait_time is None and max_batch_size has reached
    '''
    connection_str, senders = connstr_senders
    senders[0].send(EventData("Test EventData"))
    senders[1].send(EventData("Test EventData"))
    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)

    async def on_event_batch(partition_context, event_batch):
        on_event_batch.received += len(event_batch)
        await partition_context.update_checkpoint()
        on_event_batch.namespace = partition_context.fully_qualified_namespace
        on_event_batch.eventhub_name = partition_context.eventhub_name
        on_event_batch.consumer_group = partition_context.consumer_group
        on_event_batch.offset = event_batch[-1].offset
        on_event_batch.sequence_number = event_batch[-1].sequence_number

    on_event_batch.received = 0
    on_event_batch.namespace = None
    on_event_batch.eventhub_name = None
    on_event_batch.consumer_group = None
    on_event_batch.offset = None
    on_event_batch.sequence_number = None

    async with client:
        task = asyncio.ensure_future(
            client.receive_batch(on_event_batch, starting_position="-1"))
        await asyncio.sleep(10)
        assert on_event_batch.received == 2

        checkpoints = await list(client._event_processors.values())[0]._checkpoint_store.list_checkpoints(
            on_event_batch.namespace, on_event_batch.eventhub_name, on_event_batch.consumer_group
        )
        assert len([checkpoint for checkpoint in checkpoints if checkpoint["offset"] == on_event_batch.offset]) > 0
        assert len(
            [checkpoint for checkpoint in checkpoints if checkpoint["sequence_number"] == on_event_batch.sequence_number]) > 0

    await task


@pytest.mark.parametrize("max_wait_time, sleep_time, expected_result",
                         [(3, 15, []),
                          (3, 2, None),
                          ])
@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_batch_empty_with_max_wait_time_async(connection_str, max_wait_time, sleep_time, expected_result, uamqp_transport):
    '''Test whether event handler is called when max_wait_time > 0 and no event is received
    '''
    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)
    async def on_event_batch(partition_context, event_batch):
        on_event_batch.event_batch = event_batch

    on_event_batch.event_batch = None
    async with client:
        task = asyncio.ensure_future(
            client.receive_batch(on_event_batch, max_wait_time=max_wait_time, starting_position="-1"))
        await asyncio.sleep(sleep_time)
        assert on_event_batch.event_batch == expected_result

    await task


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_batch_early_callback_async(connstr_senders, uamqp_transport):
    ''' Test whether the callback is called once max_batch_size reaches and before max_wait_time reaches.
    '''
    connection_str, senders = connstr_senders
    for _ in range(10):
        senders[0].send(EventData("Test EventData"))
    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)

    async def on_event_batch(partition_context, event_batch):
        on_event_batch.received += len(event_batch)

    on_event_batch.received = 0

    async with client:
        task = asyncio.ensure_future(
            client.receive_batch(
                on_event_batch, max_batch_size=10, max_wait_time=100, starting_position="-1", partition_id="0"))
        await asyncio.sleep(10)
        assert on_event_batch.received == 10
    await task


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_batch_tracing_async(connstr_senders, uamqp_transport, fake_span):
    """Test that that receive and process spans are properly created and linked."""
    settings.tracing_implementation.set_value(fake_span)
    connection_str, senders = connstr_senders

    with fake_span(name="SendSpan") as root_send:
        senders[0].send([EventData(b"Data"), EventData(b"Data")])

    assert len(root_send.children) == 3
    assert root_send.children[0].name == "EventHubs.message"
    assert root_send.children[1].name == "EventHubs.message"
    assert root_send.children[2].name == "EventHubs.send"
    assert len(root_send.children[2].links) == 2

    traceparent1 = root_send.children[2].links[0].headers['traceparent']
    traceparent2 = root_send.children[2].links[1].headers['traceparent']

    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)

    async def on_event_batch(partition_context, event_batch):
        on_event_batch.received += len(event_batch)

    on_event_batch.received = 0

    with fake_span(name="ReceiveSpan") as root_receive:
        async with client:
            task = asyncio.ensure_future(
                client.receive_batch(on_event_batch, max_batch_size=2, starting_position="-1"))
            await asyncio.sleep(10)
            assert on_event_batch.received == 2

        await task

    assert root_receive.name == "ReceiveSpan"
    # One receive span and one process span.
    assert len(root_receive.children) == 2

    assert root_receive.children[0].name == "EventHubs.receive"
    assert root_receive.children[0].kind == SpanKind.CLIENT

    # One link for each message in the batch.
    assert len(root_receive.children[0].links) == 2
    assert root_receive.children[0].links[0].headers['traceparent'] == traceparent1
    assert root_receive.children[0].links[1].headers['traceparent'] == traceparent2

    assert root_receive.children[1].name == "EventHubs.process"
    assert root_receive.children[1].kind == SpanKind.CONSUMER

    assert len(root_receive.children[1].links) == 2
    assert root_receive.children[1].links[0].headers['traceparent'] == traceparent1
    assert root_receive.children[1].links[1].headers['traceparent'] == traceparent2

    settings.tracing_implementation.set_value(None)


@pytest.mark.liveTest
@pytest.mark.asyncio
async def test_receive_batch_large_event_async(connstr_senders, uamqp_transport):
    connection_str, senders = connstr_senders
    senders[0].send(EventData("A" * 15700))
    client = EventHubConsumerClient.from_connection_string(connection_str, consumer_group='$default', uamqp_transport=uamqp_transport)

    async def on_event(partition_context, event):
        assert partition_context.partition_id == "0"
        assert partition_context.consumer_group == "$default"
        assert partition_context.fully_qualified_namespace in connection_str
        assert partition_context.eventhub_name == senders[0]._client.eventhub_name
        on_event.received += 1
        assert client._event_processors[0]._consumers[0]._handler._link.current_link_credit == 1

    on_event.received = 0
    async with client:
        task = asyncio.ensure_future(
            client.receive(on_event, partition_id="0", starting_position="-1", prefetch=2))
        await asyncio.sleep(10)
        assert on_event.received == 1
    await task


#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------
from typing import cast
from collections import namedtuple
from enum import Enum
import struct

_AS_BYTES = struct.Struct('>B')

#: The IANA assigned port number for AMQP.The standard AMQP port number that has been assigned by IANA
#: for TCP, UDP, and SCTP.There are currently no UDP or SCTP mappings defined for AMQP.
#: The port number is reserved for future transport mappings to these protocols.
PORT = 5672

# default port for AMQP over Websocket
WEBSOCKET_PORT = 443

# subprotocol for AMQP over Websocket
AMQP_WS_SUBPROTOCOL = 'AMQPWSB10'

#: The IANA assigned port number for secure AMQP (amqps).The standard AMQP port number that has been assigned
#: by IANA for secure TCP using TLS. Implementations listening on this port should NOT expect a protocol
#: handshake before TLS is negotiated.
SECURE_PORT = 5671


# default port for AMQP over Websocket
WEBSOCKET_PORT = 443


# subprotocol for AMQP over Websocket
AMQP_WS_SUBPROTOCOL = 'AMQPWSB10'


MAJOR = 1  #: Major protocol version.
MINOR = 0  #: Minor protocol version.
REV = 0  #: Protocol revision.
HEADER_FRAME = b"AMQP\x00" + _AS_BYTES.pack(MAJOR) + _AS_BYTES.pack(MINOR) + _AS_BYTES.pack(REV)


TLS_MAJOR = 1  #: Major protocol version.
TLS_MINOR = 0  #: Minor protocol version.
TLS_REV = 0  #: Protocol revision.
TLS_HEADER_FRAME = b"AMQP\x02" + _AS_BYTES.pack(TLS_MAJOR) + _AS_BYTES.pack(TLS_MINOR) + _AS_BYTES.pack(TLS_REV)

SASL_MAJOR = 1  #: Major protocol version.
SASL_MINOR = 0  #: Minor protocol version.
SASL_REV = 0  #: Protocol revision.
SASL_HEADER_FRAME = b"AMQP\x03" + _AS_BYTES.pack(SASL_MAJOR) + _AS_BYTES.pack(SASL_MINOR) + _AS_BYTES.pack(SASL_REV)

EMPTY_FRAME = b'\x00\x00\x00\x08\x02\x00\x00\x00'

#: The lower bound for the agreed maximum frame size (in bytes). During the initial Connection negotiation, the
#: two peers must agree upon a maximum frame size. This constant defines the minimum value to which the maximum
#: frame size can be set. By defining this value, the peers can guarantee that they can send frames of up to this
#: size until they have agreed a definitive maximum frame size for that Connection.
MIN_MAX_FRAME_SIZE = 512
MAX_FRAME_SIZE_BYTES = 1024 * 1024
MAX_CHANNELS = 65535
INCOMING_WINDOW = 64 * 1024
OUTGOING_WINDOW = 64 * 1024

DEFAULT_LINK_CREDIT = 10000

FIELD = namedtuple('FIELD', 'name, type, mandatory, default, multiple')

STRING_FILTER = b"apache.org:selector-filter:string"

DEFAULT_AUTH_TIMEOUT = 60
AUTH_DEFAULT_EXPIRATION_SECONDS = 3600
TOKEN_TYPE_JWT = "jwt"
TOKEN_TYPE_SASTOKEN = "servicebus.windows.net:sastoken"
CBS_PUT_TOKEN = "put-token"
CBS_NAME = "name"
CBS_OPERATION = "operation"
CBS_TYPE = "type"
CBS_EXPIRATION = "expiration"

SEND_DISPOSITION_ACCEPT = "accepted"
SEND_DISPOSITION_REJECT = "rejected"

AUTH_TYPE_SASL_PLAIN = "AUTH_SASL_PLAIN"
AUTH_TYPE_CBS = "AUTH_CBS"

DEFAULT_WEBSOCKET_HEARTBEAT_SECONDS = 10
CONNECT_TIMEOUT = 1
SOCKET_TIMEOUT = 0.2
WS_TIMEOUT_INTERVAL = 1


class ConnectionState(Enum):
    #: In this state a Connection exists, but nothing has been sent or received. This is the state an
    #: implementation would be in immediately after performing a socket connect or socket accept.
    START = 0
    #: In this state the Connection header has been received from our peer, but we have not yet sent anything.
    HDR_RCVD = 1
    #: In this state the Connection header has been sent to our peer, but we have not yet received anything.
    HDR_SENT = 2
    #: In this state we have sent and received the Connection header, but we have not yet sent or
    #: received an open frame.
    HDR_EXCH = 3
    #: In this state we have sent both the Connection header and the open frame, but
    #: we have not yet received anything.
    OPEN_PIPE = 4
    #: In this state we have sent the Connection header, the open frame, any pipelined Connection traffic,
    #: and the close frame, but we have not yet received anything.
    OC_PIPE = 5
    #: In this state we have sent and received the Connection header, and received an open frame from
    #: our peer, but have not yet sent an open frame.
    OPEN_RCVD = 6
    #: In this state we have sent and received the Connection header, and sent an open frame to our peer,
    #: but have not yet received an open frame.
    OPEN_SENT = 7
    #: In this state we have send and received the Connection header, sent an open frame, any pipelined
    #: Connection traffic, and the close frame, but we have not yet received an open frame.
    CLOSE_PIPE = 8
    #: In this state the Connection header and the open frame have both been sent and received.
    OPENED = 9
    #: In this state we have received a close frame indicating that our partner has initiated a close.
    #: This means we will never have to read anything more from this Connection, however we can
    #: continue to write frames onto the Connection. If desired, an implementation could do a TCP half-close
    #: at this point to shutdown the read side of the Connection.
    CLOSE_RCVD = 10
    #: In this state we have sent a close frame to our partner. It is illegal to write anything more onto
    #: the Connection, however there may still be incoming frames. If desired, an implementation could do
    #: a TCP half-close at this point to shutdown the write side of the Connection.
    CLOSE_SENT = 11
    #: The DISCARDING state is a variant of the CLOSE_SENT state where the close is triggered by an error.
    #: In this case any incoming frames on the connection MUST be silently discarded until the peer's close
    #: frame is received.
    DISCARDING = 12
    #: In this state it is illegal for either endpoint to write anything more onto the Connection. The
    #: Connection may be safely closed and discarded.
    END = 13


class SessionState(Enum):
    #: In the UNMAPPED state, the Session endpoint is not mapped to any incoming or outgoing channels on the
    #: Connection endpoint. In this state an endpoint cannot send or receive frames.
    UNMAPPED = 0
    #: In the BEGIN_SENT state, the Session endpoint is assigned an outgoing channel number, but there is no entry
    #: in the incoming channel map. In this state the endpoint may send frames but cannot receive them.
    BEGIN_SENT = 1
    #: In the BEGIN_RCVD state, the Session endpoint has an entry in the incoming channel map, but has not yet
    #: been assigned an outgoing channel number. The endpoint may receive frames, but cannot send them.
    BEGIN_RCVD = 2
    #: In the MAPPED state, the Session endpoint has both an outgoing channel number and an entry in the incoming
    #: channel map. The endpoint may both send and receive frames.
    MAPPED = 3
    #: In the END_SENT state, the Session endpoint has an entry in the incoming channel map, but is no longer
    #: assigned an outgoing channel number. The endpoint may receive frames, but cannot send them.
    END_SENT = 4
    #: In the END_RCVD state, the Session endpoint is assigned an outgoing channel number, but there is no entry in
    #: the incoming channel map. The endpoint may send frames, but cannot receive them.
    END_RCVD = 5
    #: The DISCARDING state is a variant of the END_SENT state where the end is triggered by an error. In this
    #: case any incoming frames on the session MUST be silently discarded until the peer's end frame is received.
    DISCARDING = 6


class SessionTransferState(Enum):

    OKAY = 0
    ERROR = 1
    BUSY = 2


class LinkDeliverySettleReason(Enum):

    DISPOSITION_RECEIVED = 0
    SETTLED = 1
    NOT_DELIVERED = 2
    TIMEOUT = 3
    CANCELLED = 4


class LinkState(Enum):

    DETACHED = 0
    ATTACH_SENT = 1
    ATTACH_RCVD = 2
    ATTACHED = 3
    DETACH_SENT = 4
    DETACH_RCVD = 5
    ERROR = 6


class ManagementLinkState(Enum):

    IDLE = 0
    OPENING = 1
    CLOSING = 2
    OPEN = 3
    ERROR = 4


class ManagementOpenResult(Enum):

    OPENING = 0
    OK = 1
    ERROR = 2
    CANCELLED = 3


class ManagementExecuteOperationResult(Enum):

    OK = 0
    ERROR = 1
    FAILED_BAD_STATUS = 2
    LINK_CLOSED = 3


class CbsState(Enum):
    CLOSED = 0
    OPENING = 1
    OPEN = 2
    ERROR = 3


class CbsAuthState(Enum):
    OK = 0
    IDLE = 1
    IN_PROGRESS = 2
    TIMEOUT = 3
    REFRESH_REQUIRED = 4
    EXPIRED = 5
    ERROR = 6  # Put token rejected or complete but fail authentication
    FAILURE = 7  # Fail to open cbs links


class Role(object):
    """Link endpoint role.

    Valid Values:
        - False: Sender
        - True: Receiver

    <type name="role" class="restricted" source="boolean">
        <choice name="sender" value="false"/>
        <choice name="receiver" value="true"/>
    </type>
    """
    Sender = False
    Receiver = True


class SenderSettleMode(object):
    """Settlement policy for a Sender.

    Valid Values:
        - 0: The Sender will send all deliveries initially unsettled to the Receiver.
        - 1: The Sender will send all deliveries settled to the Receiver.
        - 2: The Sender may send a mixture of settled and unsettled deliveries to the Receiver.

    <type name="sender-settle-mode" class="restricted" source="ubyte">
        <choice name="unsettled" value="0"/>
        <choice name="settled" value="1"/>
        <choice name="mixed" value="2"/>
    </type>
    """
    Unsettled = 0
    Settled = 1
    Mixed = 2


class ReceiverSettleMode(object):
    """Settlement policy for a Receiver.

    Valid Values:
        - 0: The Receiver will spontaneously settle all incoming transfers.
        - 1: The Receiver will only settle after sending the disposition to the Sender and
          receiving a disposition indicating settlement of the delivery from the sender.

    <type name="receiver-settle-mode" class="restricted" source="ubyte">
        <choice name="first" value="0"/>
        <choice name="second" value="1"/>
    </type>
    """
    First = 0
    Second = 1


class SASLCode(object):
    """Codes to indicate the outcome of the sasl dialog.

    <type name="sasl-code" class="restricted" source="ubyte">
        <choice name="ok" value="0"/>
        <choice name="auth" value="1"/>
        <choice name="sys" value="2"/>
        <choice name="sys-perm" value="3"/>
        <choice name="sys-temp" value="4"/>
    </type>
    """
    #: Connection authentication succeeded.
    Ok = 0
    #: Connection authentication failed due to an unspecified problem with the supplied credentials.
    Auth = 1
    #: Connection authentication failed due to a system error.
    Sys = 2
    #: Connection authentication failed due to a system error that is unlikely to be corrected without intervention.
    SysPerm = 3
    #: Connection authentication failed due to a transient system error.
    SysTemp = 4


class MessageDeliveryState(object):

    WaitingToBeSent = 0
    WaitingForSendAck = 1
    Ok = 2
    Error = 3
    Timeout = 4
    Cancelled = 5


MESSAGE_DELIVERY_DONE_STATES = (
    MessageDeliveryState.Ok,
    MessageDeliveryState.Error,
    MessageDeliveryState.Timeout,
    MessageDeliveryState.Cancelled
)

class TransportType(Enum):
    """Transport type
    The underlying transport protocol type:
     Amqp: AMQP over the default TCP transport protocol, it uses port 5671.
     AmqpOverWebsocket: Amqp over the Web Sockets transport protocol, it uses
     port 443.
    """
    Amqp = 1
    AmqpOverWebsocket = 2

    def __eq__(self, __o: object) -> bool:
        try:
            __o = cast(Enum, __o)
            return self.value == __o.value
        except AttributeError:
            return super().__eq__(__o)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from typing import Optional, Dict, Any, Union, TYPE_CHECKING
from urllib.parse import urlparse

from azure.core.pipeline.policies import RetryMode
from ._constants import TransportType, DEFAULT_AMQPS_PORT, DEFAULT_AMQP_WSS_PORT
if TYPE_CHECKING:
    from ._transport._base import AmqpTransport
    from .aio._transport._base_async import AmqpTransportAsync



class Configuration(object):  # pylint:disable=too-many-instance-attributes
    def __init__(self, **kwargs):
        self.user_agent = kwargs.get("user_agent")  # type: Optional[str]
        self.retry_total = kwargs.get("retry_total", 3)  # type: int
        self.max_retries = self.retry_total  # type: int
        self.retry_mode = RetryMode(kwargs.get("retry_mode", "exponential"))
        self.backoff_factor = kwargs.get("retry_backoff_factor", 0.8)  # type: float
        self.backoff_max = kwargs.get("retry_backoff_max", 120)  # type: int
        self.network_tracing = kwargs.get("network_tracing", False)  # type: bool
        self.http_proxy = kwargs.get("http_proxy")  # type: Optional[Dict[str, Any]]
        self.transport_type = (
            TransportType.AmqpOverWebsocket
            if self.http_proxy
            else kwargs.get("transport_type", TransportType.Amqp)
        )
        self.auth_timeout = kwargs.get("auth_timeout", 60)  # type: int
        self.prefetch = kwargs.get("prefetch", 300)  # type: int
        self.max_batch_size = kwargs.get("max_batch_size", self.prefetch)  # type: int
        self.receive_timeout = kwargs.get("receive_timeout", 0)  # type: int
        self.send_timeout = kwargs.get("send_timeout", 60)  # type: int
        self.custom_endpoint_address = kwargs.get("custom_endpoint_address")  # type: Optional[str]
        self.connection_verify = kwargs.get("connection_verify")  # type: Optional[str]
        self.connection_port = DEFAULT_AMQPS_PORT
        self.custom_endpoint_hostname = None
        self.hostname = kwargs.pop("hostname")
        amqp_transport: Union["AmqpTransport", "AmqpTransportAsync"] = kwargs.pop("amqp_transport")
        self.socket_timeout = kwargs.get("socket_timeout", .2)

        if self.http_proxy or self.transport_type.value == TransportType.AmqpOverWebsocket.value:
            self.transport_type = TransportType.AmqpOverWebsocket
            self.connection_port = DEFAULT_AMQP_WSS_PORT
            self.socket_timeout = kwargs.get("socket_timeout", 1)
            if amqp_transport.KIND == "pyamqp":
                self.hostname += "/$servicebus/websocket"

        # custom end point
        if self.custom_endpoint_address:
            # if the custom_endpoint_address doesn't include the schema,
            # we prepend a default one to make urlparse work
            if self.custom_endpoint_address.find("//") == -1:
                self.custom_endpoint_address = "sb://" + self.custom_endpoint_address
            endpoint = urlparse(self.custom_endpoint_address)
            self.transport_type = TransportType.AmqpOverWebsocket
            self.custom_endpoint_hostname = endpoint.hostname
            if amqp_transport.KIND == "pyamqp":
                self.custom_endpoint_address += "/$servicebus/websocket"
            # in case proxy and custom endpoint are both provided, we default port to 443 if it's not provided
            self.connection_port = endpoint.port or DEFAULT_AMQP_WSS_PORT



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------
from typing import (
    Any,
    Optional,
)

class DictMixin(object):
    def __setitem__(self, key, item):
        # type: (Any, Any) -> None
        self.__dict__[key] = item

    def __getitem__(self, key):
        # type: (Any) -> Any
        return self.__dict__[key]

    def __repr__(self):
        # type: () -> str
        return str(self)

    def __len__(self):
        # type: () -> int
        return len(self.keys())

    def __delitem__(self, key):
        # type: (Any) -> None
        self.__dict__[key] = None

    def __eq__(self, other):
        # type: (Any) -> bool
        """Compare objects by comparing all attributes.

        :param any other: Any object
        :return: True if objects are equal, False otherwise.
        :rtype: bool
        """
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        # type: (Any) -> bool
        """Compare objects by comparing all attributes.

        :param any other: Any object
        :return: True if objects are not equal, False otherwise.
        :rtype: bool
        """
        return not self.__eq__(other)

    def __contains__(self, key):
        # type: (Any) -> bool
        return key in self.__dict__

    def __str__(self):
        # type: () -> str
        return str({k: v for k, v in self.__dict__.items() if not k.startswith("_")})

    def has_key(self, k):
        # type: (Any) -> bool
        return k in self.__dict__

    def update(self, *args, **kwargs):
        # type: (Any, Any) -> None
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        # type: () -> list
        return [k for k in self.__dict__ if not k.startswith("_")]

    def values(self):
        # type: () -> list
        return [v for k, v in self.__dict__.items() if not k.startswith("_")]

    def items(self):
        # type: () -> list
        return [(k, v) for k, v in self.__dict__.items() if not k.startswith("_")]

    def get(self, key, default=None):
        # type: (Any, Optional[Any]) -> Any
        if key in self.__dict__:
            return self.__dict__[key]
        return default



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import pytest
import time
from azure.identity import EnvironmentCredential, DefaultAzureCredential
from azure.eventhub import EventHubProducerClient, EventHubSharedKeyCredential
from azure.eventhub._client_base import EventHubSASTokenCredential
from azure.core.credentials import AzureSasCredential, AzureNamedKeyCredential

@pytest.mark.livetest
def test_mgmt_call_conn_str(connstr_receivers):
    connection_str, receivers = connstr_receivers
    client = EventHubProducerClient.from_connection_string(connection_str)
    client._start_producer("0",60)

@pytest.mark.livetest
def test_mgmt_call_default_azure_credential(live_eventhub):
    credential = DefaultAzureCredential()
    client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential,
                                             user_agent='customized information')
    client._start_producer("0",60)

@pytest.mark.livetest
def test_mgmt_call_credential(live_eventhub):
    credential = EnvironmentCredential()
    client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential,
                                             user_agent='customized information')
    client._start_producer("0",60)

@pytest.mark.livetest
def test_mgmt_call_sas(live_eventhub):
    hostname = live_eventhub["hostname"]
    credential = EventHubSharedKeyCredential(live_eventhub['key_name'], live_eventhub['access_key'])
    auth_uri = "sb://{}/{}".format(hostname, live_eventhub['event_hub'])
    token = credential.get_token(auth_uri).token
    client = EventHubProducerClient(fully_qualified_namespace=hostname,
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=EventHubSASTokenCredential(token, time.time() + 3000))
    client._start_producer("0",60)
    assert True

@pytest.mark.livetest
def test_mgmt_call_sas_credential(live_eventhub):
    hostname = live_eventhub["hostname"]
    credential = EventHubSharedKeyCredential(live_eventhub['key_name'], live_eventhub['access_key'])
    auth_uri = "sb://{}/{}".format(hostname, live_eventhub['event_hub'])
    token = credential.get_token(auth_uri).token
    client = EventHubProducerClient(fully_qualified_namespace=hostname,
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=AzureSasCredential(token))
    client._start_producer("0",60)
    assert True

@pytest.mark.livetest
def test_mgmt_call_azure_named_key_credential(live_eventhub):
    credential = AzureNamedKeyCredential(live_eventhub['key_name'], live_eventhub['access_key'])
    client = EventHubProducerClient(fully_qualified_namespace=live_eventhub['hostname'],
                                             eventhub_name=live_eventhub['event_hub'],
                                             credential=credential)

    client._start_producer("0",60)
    assert True



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from an Event Hub partition with EventHubConsumerClient tracking
the last enqueued event properties of specific partition.
"""
import os
from azure.eventhub import EventHubConsumerClient

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']


def on_event(partition_context, event):
    print("Received event from partition {}.".format(partition_context.partition_id))
    
    # Put your code here. to do some operations on the event.
    # Avoid time-consuming operations.
    print(event)

    print("Last enqueued event properties from partition: {} is: {}.".format(
        partition_context.partition_id,
        partition_context.last_enqueued_event_properties)
    )


if __name__ == '__main__':
    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group='$Default',
        eventhub_name=EVENTHUB_NAME,
    )

    try:
        with consumer_client:
            consumer_client.receive(
                on_event=on_event,
                partition_id='0',
                track_last_enqueued_event_properties=True,
                starting_position="-1",  # "-1" is from the beginning of the partition.
            )
    except KeyboardInterrupt:
        print('Stopped receiving.')



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------

import os
import pytest
from azure.eventhub import (
    EventHubConnectionStringProperties,
    parse_connection_string,
)


class TestEventHubConnectionStringParser:

    def test_eh_conn_str_parse_cs(self, **kwargs):
        conn_str = 'Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == 'sb://eh-namespace.servicebus.windows.net/'
        assert parse_result.fully_qualified_namespace == 'eh-namespace.servicebus.windows.net'
        assert parse_result.shared_access_key_name == 'test-policy'
        assert parse_result.shared_access_key == 'THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='

    def test_eh_conn_str_parse_with_entity_path(self, **kwargs):
        conn_str = 'Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;EntityPath=eventhub-name'
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == 'sb://eh-namespace.servicebus.windows.net/'
        assert parse_result.fully_qualified_namespace == 'eh-namespace.servicebus.windows.net'
        assert parse_result.shared_access_key_name == 'test-policy'
        assert parse_result.shared_access_key == 'THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        assert parse_result.eventhub_name == 'eventhub-name'

    def test_eh_conn_str_parse_sas_and_shared_key(self, **kwargs):
        conn_str = 'Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;SharedAccessSignature=THISISASASXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == 'Only one of the SharedAccessKey or SharedAccessSignature must be present.'

    def test_eh_parse_malformed_conn_str_no_endpoint(self, **kwargs):
        conn_str = 'SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == 'Connection string is either blank or malformed.'

    def test_eh_parse_malformed_conn_str_no_endpoint_value(self, **kwargs):
        conn_str = 'Endpoint=;SharedAccessKeyName=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == 'Connection string is either blank or malformed.'

    def test_eh_parse_malformed_conn_str_no_netloc(self, **kwargs):
        conn_str = 'Endpoint=MALFORMED;SharedAccessKeyName=test-policy;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == 'Invalid Endpoint on the Connection String.'

    def test_eh_parse_conn_str_sas(self, **kwargs):
        conn_str = 'Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessSignature=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == 'sb://eh-namespace.servicebus.windows.net/'
        assert parse_result.fully_qualified_namespace == 'eh-namespace.servicebus.windows.net'
        assert parse_result.shared_access_signature == 'THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        assert parse_result.shared_access_key_name == None

    def test_eh_parse_conn_str_whitespace_trailing_semicolon(self, **kwargs):
        conn_str = '    Endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessSignature=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;    '
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == 'sb://resourcename.servicebus.windows.net/'
        assert parse_result.fully_qualified_namespace == 'resourcename.servicebus.windows.net'
        assert parse_result.shared_access_signature == 'THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        assert parse_result.shared_access_key_name == None

    def test_eh_parse_conn_str_sas_trailing_semicolon(self, **kwargs):
        conn_str = 'Endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessSignature=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX=;'
        parse_result = parse_connection_string(conn_str)
        assert parse_result.endpoint == 'sb://resourcename.servicebus.windows.net/'
        assert parse_result.fully_qualified_namespace == 'resourcename.servicebus.windows.net'
        assert parse_result.shared_access_signature == 'THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        assert parse_result.shared_access_key_name == None

    def test_eh_parse_conn_str_no_keyname(self, **kwargs):
        conn_str = 'Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_parse_conn_str_no_key(self, **kwargs):
        conn_str = 'Endpoint=sb://eh-namespace.servicebus.windows.net/;SharedAccessKeyName=test-policy'
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_parse_conn_str_no_key_or_sas(self, **kwargs):
        conn_str = 'Endpoint=sb://resourcename.servicebus.windows.net/'
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == 'At least one of the SharedAccessKey or SharedAccessSignature must be present.'

    def test_eh_parse_malformed_conn_str_lowercase_endpoint(self, **kwargs):
        conn_str = 'endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessKeyName=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert str(e.value) == 'Connection string is either blank or malformed.'

    def test_eh_parse_malformed_conn_str_lowercase_sa_key_name(self, **kwargs):
        conn_str = 'Endpoint=sb://resourcename.servicebus.windows.net/;sharedaccesskeyname=test;SharedAccessKey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)

    def test_eh_parse_malformed_conn_str_lowercase_sa_key_name(self, **kwargs):
        conn_str = 'Endpoint=sb://resourcename.servicebus.windows.net/;SharedAccessKeyName=test;sharedaccesskey=THISISATESTKEYXXXXXXXXXXXXXXXXXXXXXXXXXXXX='
        with pytest.raises(ValueError) as e:
            parse_result = parse_connection_string(conn_str)
        assert "Invalid connection string" in str(e.value)



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to show streaming sending events with different options to an Event Hub.
"""

import time
import os

from azure.eventhub import EventHubProducerClient, EventData

CONNECTION_STR = os.environ['EVENT_HUB_CONN_STR']
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']


start_time = time.time()

producer = EventHubProducerClient.from_connection_string(
    conn_str=CONNECTION_STR,
    eventhub_name=EVENTHUB_NAME
)
to_send_message_cnt = 500
bytes_per_message = 256

with producer:
    event_data_batch = producer.create_batch()
    for i in range(to_send_message_cnt):
        event_data = EventData('D' * bytes_per_message)
        try:
            event_data_batch.add(event_data)
        except ValueError:
            producer.send_batch(event_data_batch)
            event_data_batch = producer.create_batch()
            event_data_batch.add(event_data)
    if len(event_data_batch) > 0:
        producer.send_batch(event_data_batch)

print("Send messages in {} seconds.".format(time.time() - start_time))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations
from datetime import datetime
from typing import (
    Union,
    List,
    Dict,
    Callable,
    Any,
    Tuple,
    Optional,
    cast,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports
    from .._common import EventData
    from .._consumer import EventHubConsumer
    from ..aio._consumer_async import (
        EventHubConsumer as EventHubConsumerAsync
    )
    from .._consumer_client import EventHubConsumerClient
    from ..aio._consumer_client_async import (
        EventHubConsumerClient as EventHubConsumerClientAsync,
    )


class EventProcessorMixin(object):

    _eventhub_client: Optional[Union[EventHubConsumerClient, EventHubConsumerClientAsync]] = None
    _consumer_group = ""  # type: str
    _owner_level = None  # type: Optional[int]
    _prefetch = None  # type: Optional[int]
    _track_last_enqueued_event_properties = False  # type: bool
    _initial_event_position_inclusive = {}  # type: Union[bool, Dict[str, bool]]
    _initial_event_position = (
        {}
    )  # type: Union[int, str, datetime, Dict[str, Union[int, str, datetime]]]

    def get_init_event_position(
        self,
        partition_id: str,
        checkpoint: Optional[Dict[str, Any]]
        ) -> Tuple[Union[str, int, datetime], bool]:
        checkpoint_offset = checkpoint.get("offset") if checkpoint else None

        event_position_inclusive = False
        if isinstance(self._initial_event_position_inclusive, dict):
            event_position_inclusive = self._initial_event_position_inclusive.get(
                partition_id, False
            )
        elif isinstance(self._initial_event_position_inclusive, bool):
            event_position_inclusive = self._initial_event_position_inclusive

        event_position = "-1"  # type: Union[int, str, datetime]
        if checkpoint_offset:
            event_position = checkpoint_offset
        elif isinstance(self._initial_event_position, dict):
            event_position = self._initial_event_position.get(partition_id, "-1")  # type: ignore
        else:
            event_position = cast(
                Union[int, str, datetime], self._initial_event_position
            )

        return event_position, event_position_inclusive

    def create_consumer(
        self,
        partition_id: str,
        initial_event_position: Union[str, int, datetime],
        initial_event_position_inclusive: bool,
        on_event_received: Callable[[Union[Optional[EventData], List[EventData]]], None],
        **kwargs: Any
    ) -> Union[EventHubConsumer, EventHubConsumerAsync]:
        consumer = self._eventhub_client._create_consumer(  # type: ignore  # pylint: disable=protected-access
            self._consumer_group,
            partition_id,
            initial_event_position,
            on_event_received,  # type: ignore
            event_position_inclusive=initial_event_position_inclusive,
            owner_level=self._owner_level,
            track_last_enqueued_event_properties=self._track_last_enqueued_event_properties,
            prefetch=self._prefetch,
            **kwargs,
        )
        return consumer



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------
import sys
import os
import pytest
import logging
import uuid
import warnings
import datetime
from logging.handlers import RotatingFileHandler

from azure.identity import EnvironmentCredential
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.eventhub import EventHubManagementClient
from azure.eventhub import EventHubProducerClient
from azure.eventhub._pyamqp import ReceiveClient
from azure.eventhub._pyamqp.authentication import SASTokenAuth
from azure.eventhub.extensions.checkpointstoreblob import BlobCheckpointStore
from azure.eventhub.extensions.checkpointstoreblobaio import BlobCheckpointStore as BlobCheckpointStoreAsync
try:
    import uamqp
    uamqp_transport_params = [True, False]
    uamqp_transport_ids = ["uamqp", "pyamqp"]
except (ModuleNotFoundError, ImportError):
    uamqp_transport_params = [False]
    uamqp_transport_ids = ["pyamqp"]

from devtools_testutils import get_region_override
from tracing_common import FakeSpan

collect_ignore = []
PARTITION_COUNT = 2
CONN_STR = "Endpoint=sb://{}/;SharedAccessKeyName={};SharedAccessKey={};EntityPath={}"
RES_GROUP_PREFIX = "eh-res-group"
NAMESPACE_PREFIX = "eh-ns"
EVENTHUB_PREFIX = "eh"
EVENTHUB_DEFAULT_AUTH_RULE_NAME = 'RootManageSharedAccessKey'
LOCATION = get_region_override("westus")


def pytest_addoption(parser):
    parser.addoption(
        "--sleep", action="store", default="True", help="sleep on reconnect test: True or False"
    )


@pytest.fixture
def sleep(request):
    sleep = request.config.getoption("--sleep")
    return sleep.lower() in ('true', 'yes', '1', 'y')

@pytest.fixture(scope="session", params=uamqp_transport_params, ids=uamqp_transport_ids)
def uamqp_transport(request):
    return request.param

@pytest.fixture(scope="session")    
def storage_connection_str():
    try:
        return os.environ['AZURE_STORAGE_CONN_STR']
    except KeyError:
        pytest.skip('AZURE_STORAGE_CONN_STR undefined')
        return

@pytest.fixture()    
def checkpoint_store(storage_connection_str):
    checkpoint_store = BlobCheckpointStore.from_connection_string(storage_connection_str, "blobcontainer" + str(uuid.uuid4()))
    return checkpoint_store

@pytest.fixture()    
def checkpoint_store_aio(storage_connection_str):
    checkpoint_store = BlobCheckpointStoreAsync.from_connection_string(storage_connection_str, "blobcontainer" + str(uuid.uuid4()))
    return checkpoint_store

def get_logger(filename, level=logging.INFO):
    azure_logger = logging.getLogger("azure.eventhub")
    azure_logger.setLevel(level)
    uamqp_logger = logging.getLogger("uamqp")
    uamqp_logger.setLevel(logging.INFO)

    formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    console_handler = logging.StreamHandler(stream=sys.stdout)
    console_handler.setFormatter(formatter)
    if not azure_logger.handlers:
        azure_logger.addHandler(console_handler)
    if not uamqp_logger.handlers:
        uamqp_logger.addHandler(console_handler)

    if filename:
        file_handler = RotatingFileHandler(filename, maxBytes=5*1024*1024, backupCount=2)
        file_handler.setFormatter(formatter)
        azure_logger.addHandler(file_handler)

    return azure_logger


log = get_logger(None, logging.DEBUG)


@pytest.fixture(scope="session")
def timeout_factor(uamqp_transport):
    if uamqp_transport:
        return 1000
    else:
        return 1

@pytest.fixture(scope="session")
def fake_span():
    return FakeSpan

@pytest.fixture(scope="session")
def resource_group():
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        pytest.skip('AZURE_SUBSCRIPTION_ID undefined')
        return
    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = ResourceManagementClient(EnvironmentCredential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes)
    resource_group_name = RES_GROUP_PREFIX + str(uuid.uuid4())
    parameters = {"location": LOCATION}
    expiry = datetime.datetime.utcnow() + datetime.timedelta(days=1)
    parameters['tags'] = {'DeleteAfter': expiry.replace(microsecond=0).isoformat()}
    try:
        rg = resource_client.resource_groups.create_or_update(
            resource_group_name,
            parameters
        )
        yield rg
    finally:
        try:
            resource_client.resource_groups.begin_delete(resource_group_name)
        except:
            warnings.warn(UserWarning("resource group teardown failed"))


@pytest.fixture(scope="session")
def eventhub_namespace(resource_group):
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        pytest.skip('AZURE_SUBSCRIPTION_ID defined')
        return
    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(EnvironmentCredential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes)
    namespace_name = NAMESPACE_PREFIX + str(uuid.uuid4())
    try:
        namespace = resource_client.namespaces.begin_create_or_update(
            resource_group.name, namespace_name, {"location": LOCATION}
        ).result()
        key = resource_client.namespaces.list_keys(resource_group.name, namespace_name, EVENTHUB_DEFAULT_AUTH_RULE_NAME)
        connection_string = key.primary_connection_string
        key_name = key.key_name
        primary_key = key.primary_key
        yield namespace.name, connection_string, key_name, primary_key
    finally:
        try:
            resource_client.namespaces.begin_delete(resource_group.name, namespace_name).wait()
        except:
            warnings.warn(UserWarning("eventhub namespace teardown failed"))


@pytest.fixture()
def live_eventhub(resource_group, eventhub_namespace):  # pylint: disable=redefined-outer-name
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        pytest.skip('AZURE_SUBSCRIPTION_ID defined')
        return
    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(EnvironmentCredential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes)
    eventhub_name = EVENTHUB_PREFIX + str(uuid.uuid4())
    eventhub_ns_name, connection_string, key_name, primary_key = eventhub_namespace
    eventhub_endpoint_suffix = os.environ.get("EVENT_HUB_ENDPOINT_SUFFIX", ".servicebus.windows.net")
    try:
        eventhub = resource_client.event_hubs.create_or_update(
            resource_group.name, eventhub_ns_name, eventhub_name, {"partition_count": PARTITION_COUNT}
        )
        live_eventhub_config = {
            'resource_group': resource_group.name,
            'hostname': "{}{}".format(eventhub_ns_name, eventhub_endpoint_suffix),
            'key_name': key_name,
            'access_key': primary_key,
            'namespace': eventhub_ns_name,
            'event_hub': eventhub.name,
            'consumer_group': '$Default',
            'partition': '0',
            'connection_str': connection_string + ";EntityPath="+eventhub.name
        }
        yield live_eventhub_config
    finally:
        try:
            resource_client.event_hubs.delete(resource_group.name, eventhub_ns_name, eventhub_name)
        except:
            warnings.warn(UserWarning("eventhub teardown failed"))

@pytest.fixture()
def resource_mgmt_client():
    try:
        SUBSCRIPTION_ID = os.environ["AZURE_SUBSCRIPTION_ID"]
    except KeyError:
        pytest.skip('AZURE_SUBSCRIPTION_ID defined')
        return
    base_url = os.environ.get("EVENTHUB_RESOURCE_MANAGER_URL", "https://management.azure.com/")
    credential_scopes = ["{}.default".format(base_url)]
    resource_client = EventHubManagementClient(EnvironmentCredential(), SUBSCRIPTION_ID, base_url=base_url, credential_scopes=credential_scopes)
    yield resource_client

@pytest.fixture()
def connection_str(live_eventhub):
    return CONN_STR.format(
        live_eventhub['hostname'],
        live_eventhub['key_name'],
        live_eventhub['access_key'],
        live_eventhub['event_hub'])


@pytest.fixture()
def invalid_hostname(live_eventhub):
    return CONN_STR.format(
        "invalid123.servicebus.windows.net",
        live_eventhub['key_name'],
        live_eventhub['access_key'],
        live_eventhub['event_hub'])


@pytest.fixture()
def invalid_key(live_eventhub):
    return CONN_STR.format(
        live_eventhub['hostname'],
        live_eventhub['key_name'],
        "invalid",
        live_eventhub['event_hub'])


@pytest.fixture()
def invalid_policy(live_eventhub):
    return CONN_STR.format(
        live_eventhub['hostname'],
        "invalid",
        live_eventhub['access_key'],
        live_eventhub['event_hub'])


@pytest.fixture()
def connstr_receivers(live_eventhub, uamqp_transport):
    connection_str = live_eventhub["connection_str"]
    partitions = [str(i) for i in range(PARTITION_COUNT)]
    receivers = []
    for p in partitions:
        uri = "sb://{}/{}".format(live_eventhub['hostname'], live_eventhub['event_hub'])
        source = "amqps://{}/{}/ConsumerGroups/{}/Partitions/{}".format(
            live_eventhub['hostname'],
            live_eventhub['event_hub'],
            live_eventhub['consumer_group'],
            p)
        if uamqp_transport:
            sas_auth = uamqp.authentication.SASTokenAuth.from_shared_access_key(
                uri, live_eventhub['key_name'], live_eventhub['access_key'])
            receiver = uamqp.ReceiveClient(source, auth=sas_auth, debug=False, timeout=0, prefetch=500)
        else:
            sas_auth = SASTokenAuth(
                uri, uri, live_eventhub['key_name'], live_eventhub['access_key']
            )
            receiver = ReceiveClient(live_eventhub['hostname'], source, auth=sas_auth, network_trace=False, timeout=0, link_credit=500)
        receiver.open()
        receivers.append(receiver)
    yield connection_str, receivers
    for r in receivers:
        r.close()


@pytest.fixture()
def connstr_senders(live_eventhub, uamqp_transport):
    connection_str = live_eventhub["connection_str"]
    client = EventHubProducerClient.from_connection_string(connection_str, uamqp_transport=uamqp_transport)
    partitions = client.get_partition_ids()

    senders = []
    for p in partitions:
        sender = client._create_producer(partition_id=p)
        senders.append(sender)
    yield connection_str, senders
    for s in senders:
        s.close()
    client.close()

# Note: This is duplicated between here and the basic conftest, so that it does not throw warnings if you're
# running locally to this SDK. (Everything works properly, pytest just makes a bit of noise.)
def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers", "liveTest: mark test to be a live test only"
    )



import asyncio
import sys

if sys.version_info >= (3, 8, 0):
    from unittest.mock import AsyncMock
from azure.eventhub._pyamqp.aio import Link
from azure.eventhub._pyamqp.constants import LinkState
import pytest


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "start_state,expected_state",
    [
        (LinkState.ATTACHED, LinkState.DETACH_SENT),
        (LinkState.ATTACH_SENT, LinkState.DETACHED),
        (LinkState.ATTACH_RCVD, LinkState.DETACHED),
    ],
    ids=["link attached", "link attach sent", "link attach rcvd"],
)
async def test_link_should_detach(start_state, expected_state):
    if sys.version_info < (3, 8, 0):
        pytest.skip("AsyncMock is not available in Python 3.7")
    session = AsyncMock()
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    await link._set_state(start_state)
    link._outgoing_detach = AsyncMock(return_value=None)
    await link.detach()

    assert link.state == expected_state


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "state",
    [LinkState.DETACHED, LinkState.DETACH_SENT, LinkState.ERROR],
    ids=["link detached", "link detach sent", "link error"],
)
async def test_link_should_not_detach(state):
    if sys.version_info < (3, 8, 0):
        pytest.skip("AsyncMock is not available in Python 3.7")
    session = None
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    await link._set_state(state)
    link._outgoing_detach = AsyncMock(return_value=None)
    await link.detach()
    link._outgoing_detach.assert_not_called()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# -------------------------------------------------------------------------

from ._amqp_message import (
    AmqpAnnotatedMessage,
    AmqpMessageBodyType,
    AmqpMessageProperties,
    AmqpMessageHeader,
)


__all__ = [
    "AmqpAnnotatedMessage",
    "AmqpMessageBodyType",
    "AmqpMessageProperties",
    "AmqpMessageHeader",
]



#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#-------------------------------------------------------------------------
from functools import partial

from ..authentication import (
    _generate_sas_access_token,
    SASTokenAuth,
    JWTTokenAuth
)
from ..constants import AUTH_DEFAULT_EXPIRATION_SECONDS


async def _generate_sas_token_async(auth_uri, sas_name, sas_key, expiry_in=AUTH_DEFAULT_EXPIRATION_SECONDS):
    return _generate_sas_access_token(auth_uri, sas_name, sas_key, expiry_in=expiry_in)


class JWTTokenAuthAsync(JWTTokenAuth):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    ...


class SASTokenAuthAsync(SASTokenAuth):
    # TODO:
    #  1. naming decision, suffix with Auth vs Credential
    def __init__(
        self,
        uri,
        audience,
        username,
        password,
        **kwargs
    ):
        """
        CBS authentication using SAS tokens.

        :param uri: The AMQP endpoint URI. This must be provided as
         a decoded string.
        :type uri: str
        :param audience: The token audience field. For SAS tokens
         this is usually the URI.
        :type audience: str
        :param username: The SAS token username, also referred to as the key
         name or policy name. This can optionally be encoded into the URI.
        :type username: str
        :param password: The SAS token password, also referred to as the key.
         This can optionally be encoded into the URI.
        :type password: str
        :param expires_in: The total remaining seconds until the token
         expires.
        :type expires_in: int
        :param expires_on: The timestamp at which the SAS token will expire
         formatted as seconds since epoch.
        :type expires_on: float
        :param token_type: The type field of the token request.
         Default value is `"servicebus.windows.net:sastoken"`.
        :type token_type: str

        """
        super(SASTokenAuthAsync, self).__init__(
            uri,
            audience,
            username,
            password,
            **kwargs
        )
        self.get_token = partial(_generate_sas_token_async, uri, username, password, self.expires_in)



from unittest.mock import Mock
from azure.eventhub._pyamqp.link import Link
from azure.eventhub._pyamqp.receiver import ReceiverLink
from azure.eventhub._pyamqp.constants import LinkState
import pytest


@pytest.mark.parametrize(
    "start_state,expected_state",
    [
        (LinkState.ATTACHED, LinkState.DETACH_SENT),
        (LinkState.ATTACH_SENT, LinkState.DETACHED),
        (LinkState.ATTACH_RCVD, LinkState.DETACHED),
    ],
    ids=["link attached", "link attach sent", "link attach rcvd"],
)
def test_link_should_detach(start_state, expected_state):
    session = Mock()
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    link._set_state(start_state)
    link._outgoing_detach = Mock(return_value=None)
    link.detach()

    assert link.state == expected_state


@pytest.mark.parametrize(
    "state",
    [LinkState.DETACHED, LinkState.DETACH_SENT, LinkState.ERROR],
    ids=["link detached", "link detach sent", "link error"],
)
def test_link_should_not_detach(state):
    session = None
    link = Link(
        session,
        3,
        name="test_link",
        role=True,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
    )
    assert link.state == LinkState.DETACHED

    link._set_state(state)
    link._outgoing_detach = Mock(return_value=None)
    link.detach()
    link._outgoing_detach.assert_not_called()

def test_receive_transfer_frame_multiple():
    session = None
    link = ReceiverLink(
        session,
        3,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
        on_transfer=Mock(),
    )

    link.current_link_credit = 2 # Set the link credit to 2

    # frame: handle, delivery_id, delivery_tag, message_format, settled, more, rcv_settle_mode, state, resume, aborted, bathable, payload
    transfer_frame_one = [3, 0, b'/blah', 0, True, True, None, None, None, None, False, ""]
    transfer_frame_two = [3, None, b'/blah', 0, True, False, None, None, None, None, False, ""]

    link._incoming_transfer(transfer_frame_one)
    assert link.current_link_credit == 2
    link._incoming_transfer(transfer_frame_two)
    assert link.current_link_credit == 1

def test_receive_transfer_continuation_frame():
    session = None
    link = ReceiverLink(
        session,
        3,
        source_address="test_source",
        target_address="test_target",
        network_trace=False,
        network_trace_params={},
        on_transfer=Mock(),
    )

    link.current_link_credit = 3 # Set the link credit to 2

    # frame: handle, delivery_id, delivery_tag, message_format, settled, more, rcv_settle_mode, state, resume, aborted, batchable, payload
    transfer_frame_one = [3, 0, b'/blah', 0, True, False, None, None, None, None, False, ""]
    transfer_frame_two = [3, 1, b'/blah', 0, True, True, None, None, None, None, False, ""]
    transfer_frame_three = [3, None, b'/blah', 0, True, False, None, None, None, None, False, ""]



    link._incoming_transfer(transfer_frame_one)
    assert link.current_link_credit == 2
    link._incoming_transfer(transfer_frame_two)
    assert link.current_link_credit == 2
    link._incoming_transfer(transfer_frame_three)
    assert link.current_link_credit == 1


#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
Example to demonstrate utilizing azure.core.credentials.AzureSasCredential to authenticate with Event Hubs.
"""

import os
import time
import hmac
import hashlib
import base64
from urllib.parse import quote as url_parse_quote

from azure.core.credentials import AzureSasCredential
from azure.eventhub import EventHubProducerClient, EventData


def generate_sas_token(uri, sas_name, sas_value, token_ttl):
    """Performs the signing and encoding needed to generate a sas token from a sas key."""
    sas = sas_value.encode('utf-8')
    expiry = str(int(time.time() + token_ttl))
    string_to_sign = (uri + '\n' + expiry).encode('utf-8')
    signed_hmac_sha256 = hmac.HMAC(sas, string_to_sign, hashlib.sha256)
    signature = url_parse_quote(base64.b64encode(signed_hmac_sha256.digest()))
    return 'SharedAccessSignature sr={}&sig={}&se={}&skn={}'.format(uri, signature, expiry, sas_name)


# Target namespace and hub must also be specified.  Consumer group is set to default unless required otherwise.
FULLY_QUALIFIED_NAMESPACE = os.environ['EVENT_HUB_HOSTNAME']
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']

# The following part creates a SAS token. A SAS token is typically given to you after being created.
SAS_POLICY = os.environ['EVENT_HUB_SAS_POLICY']
SAS_KEY = os.environ['EVENT_HUB_SAS_KEY']

uri = "sb://{}/{}".format(FULLY_QUALIFIED_NAMESPACE, EVENTHUB_NAME)
token_ttl = 3000  # seconds
sas_token = generate_sas_token(uri, SAS_POLICY, SAS_KEY, token_ttl)

credential = AzureSasCredential(sas_token)

producer_client = EventHubProducerClient(
    fully_qualified_namespace=FULLY_QUALIFIED_NAMESPACE,
    eventhub_name=EVENTHUB_NAME,
    credential=credential,
    logging_enable=True
)

start_time = time.time()
with producer_client:
    event_data_batch = producer_client.create_batch()
    event_data_batch.add(EventData('Single message'))
    producer_client.send_batch(event_data_batch)

print("Send messages in {} seconds.".format(time.time() - start_time))



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from __future__ import annotations
from contextlib import contextmanager
from enum import Enum
import logging
from typing import (
    TYPE_CHECKING,
    Type,
    Optional,
    Dict,
    List,
    Union,
    Any,
    Iterable,
    Iterator
)

from azure.core import CaseInsensitiveEnumMeta
from azure.core.settings import settings
from azure.core.tracing import SpanKind, Link

from .amqp import AmqpAnnotatedMessage
from ._constants import PROP_TIMESTAMP

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports
    from ._transport._base import AmqpTransport
    try:
        from uamqp import Message as uamqp_Message
    except ImportError:
        uamqp_Message = None
    from ._pyamqp.message import Message
    from azure.core.tracing import AbstractSpan
    from ._client_base import ClientBase
    from ._common import EventData, EventDataBatch


_LOGGER = logging.getLogger(__name__)


TRACE_DIAGNOSTIC_ID_PROPERTY = b"Diagnostic-Id"
TRACE_PARENT_PROPERTY = b"traceparent"
TRACE_STATE_PROPERTY = b"tracestate"
TRACE_PROPERTY_ENCODING = "ascii"


class TraceAttributes:
    TRACE_NAMESPACE_ATTRIBUTE = "az.namespace"
    TRACE_NAMESPACE = "Microsoft.EventHub"

    TRACE_MESSAGING_SYSTEM_ATTRIBUTE = "messaging.system"
    TRACE_MESSAGING_SYSTEM = "eventhubs"

    TRACE_NET_PEER_NAME_ATTRIBUTE = "net.peer.name"
    TRACE_MESSAGING_DESTINATION_ATTRIBUTE = "messaging.destination.name"
    TRACE_MESSAGING_OPERATION_ATTRIBUTE = "messaging.operation"
    TRACE_MESSAGING_BATCH_COUNT_ATTRIBUTE = "messaging.batch.message_count"

    LEGACY_TRACE_MESSAGE_BUS_DESTINATION_ATTRIBUTE = "message_bus.destination"
    LEGACY_TRACE_PEER_ADDRESS_ATTRIBUTE = "peer.address"


class TraceOperationTypes(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    PUBLISH = "publish"
    RECEIVE = "receive"
    PROCESS = "process"


def is_tracing_enabled():
    span_impl_type = settings.tracing_implementation()
    return span_impl_type is not None


@contextmanager
def send_context_manager(client: Optional[ClientBase], links: Optional[List[Link]] = None) -> Iterator[None]:
    """Tracing for message sending.

    :param ~azure.eventhub._client_base.ClientBase or None client: The client that is sending the message.
    :param list[~azure.core.tracing.Link] or None links: A list of links to add to the span.
    :return: A context manager that will start and end the span.
    :rtype: iterator[None]
    """
    span_impl_type: Optional[Type[AbstractSpan]] = settings.tracing_implementation()
    if span_impl_type is not None:
        links = links or []
        with span_impl_type(name="EventHubs.send", kind=SpanKind.CLIENT, links=links) as span:
            add_span_attributes(span, TraceOperationTypes.PUBLISH, client, message_count=len(links))
            yield
    else:
        yield


@contextmanager
def receive_context_manager(
    client: Optional[ClientBase], links: Optional[List[Link]] = None, start_time: Optional[int] = None
)  -> Iterator[None]:
    """Tracing for message receiving.

    :param ~azure.eventhub._client_base.ClientBase or None client: The client that is receiving the message.
    :param list[~azure.core.tracing.Link] or None links: A list of links to add to the span.
    :param int or None start_time: The time the receive operation started.
    :return: A context manager that will start and end the span.
    :rtype: iterator[None]
    """
    span_impl_type: Optional[Type[AbstractSpan]] = settings.tracing_implementation()
    if span_impl_type is not None:
        links = links or []
        with span_impl_type(
            name="EventHubs.receive", kind=SpanKind.CLIENT, links=links, start_time=start_time
        ) as span:
            add_span_attributes(span, TraceOperationTypes.RECEIVE, client, message_count=len(links))
            yield
    else:
        yield


@contextmanager
def process_context_manager(
    client: Optional[ClientBase], links: Optional[List[Link]] = None, is_batch: bool = False
) -> Iterator[None]:
    """Tracing for message processing.

    :param ~azure.eventhub._client_base.ClientBase or None client: The client that is processing the message.
    :param list[~azure.core.tracing.Link] or None links: A list of links to add to the span.
    :param bool is_batch: Whether the processing is done in a batch.
    :return: A context manager that will start and end the span.
    :rtype: iterator[None]
    """
    span_impl_type: Optional[Type[AbstractSpan]] = settings.tracing_implementation()
    if span_impl_type is not None:
        context = None
        links = links or []

        # If the processing callback is called per single message, the processing span should be a child of the
        # context of the message (as opposed to messages being links in the processing span).
        if not is_batch and links:
            context = links[0].headers
            links = []
        with span_impl_type(name="EventHubs.process", kind=SpanKind.CONSUMER, links=links, context=context) as span:
            add_span_attributes(span, TraceOperationTypes.PROCESS, client, message_count=len(links))
            yield
    else:
        yield


def trace_message(
    message: Union[uamqp_Message, Message],
    amqp_transport: AmqpTransport,
    additional_attributes: Optional[Dict[str, Any]] = None
) -> Union[uamqp_Message, Message]:
    """Add tracing information to the message and return the updated message.

    Will open and close an message span, and add trace context to the app properties of the message.

    :param uamqp.Message or ~azure.eventhub._pyamqp.message.Message message: The message to trace.
    :param ~azure.eventhub._transport._base.AmqpTransport amqp_transport: The transport to use for tracing.
    :param dict[str,any] or None additional_attributes: Additional attributes to add to the span.
    :rtype: uamqp.Message or ~azure.eventhub._pyamqp.message.Message
    :return: The message with tracing information added.
    """
    try:
        span_impl_type: Optional[Type[AbstractSpan]] = settings.tracing_implementation()
        if span_impl_type is not None:
            with span_impl_type(name="EventHubs.message", kind=SpanKind.PRODUCER) as message_span:
                headers = message_span.to_header()

                if "traceparent" in headers:
                    message = amqp_transport.update_message_app_properties(
                        message,
                        TRACE_DIAGNOSTIC_ID_PROPERTY,
                        headers["traceparent"]
                    )
                    message = amqp_transport.update_message_app_properties(
                        message,
                        TRACE_PARENT_PROPERTY,
                        headers["traceparent"]
                    )
                if "tracestate" in headers:
                    message = amqp_transport.update_message_app_properties(
                        message,
                        TRACE_STATE_PROPERTY,
                        headers["tracestate"]
                    )

                message_span.add_attribute(
                    TraceAttributes.TRACE_NAMESPACE_ATTRIBUTE, TraceAttributes.TRACE_NAMESPACE)
                message_span.add_attribute(
                    TraceAttributes.TRACE_MESSAGING_SYSTEM_ATTRIBUTE, TraceAttributes.TRACE_MESSAGING_SYSTEM)

                if additional_attributes:
                    for key, value in additional_attributes.items():
                        if value is not None:
                            message_span.add_attribute(key, value)

    except Exception as exp:  # pylint:disable=broad-except
        _LOGGER.warning("trace_message had an exception %r", exp)

    return message


def get_span_links_from_received_events(events: Union[EventData, Iterable[EventData]]) -> List[Link]:
    """Create span links from received events.

    This will extract the traceparent and tracestate from the event properties and create span links
    based on these values. The time the event was enqueued is also added as a link attribute.

    :param ~azure.eventhub.EventData or iterable[~azure.eventhub.EventData] events: The received events.
    :rtype: list[~azure.core.tracing.Link]
    :return: A list of span links.
    """
    # pylint:disable=isinstance-second-argument-not-valid-type
    trace_events = events if isinstance(events, Iterable) else (events,)
    links = []
    try:
        for event in trace_events:
            headers = {}
            if event.properties:
                traceparent = (event.properties.get(TRACE_PARENT_PROPERTY, b"") or
                               event.properties.get(TRACE_DIAGNOSTIC_ID_PROPERTY, b""))
                if hasattr(traceparent, "decode"):
                    traceparent = traceparent.decode(TRACE_PROPERTY_ENCODING)
                if traceparent:
                    headers["traceparent"] = traceparent

                tracestate = event.properties.get(TRACE_STATE_PROPERTY, b"")
                if hasattr(tracestate, "decode"):
                    tracestate = tracestate.decode(TRACE_PROPERTY_ENCODING)
                if tracestate:
                    headers["tracestate"] = tracestate

            enqueued_time = event.system_properties.get(PROP_TIMESTAMP)
            attributes = {'enqueuedTime': enqueued_time} if enqueued_time else None
            links.append(Link(headers, attributes=attributes))
    except AttributeError:
        pass
    return links


def get_span_links_from_batch(batch: EventDataBatch) -> List[Link]:
    """Create span links from a batch of events.

    :param ~azure.eventhub.EventDataBatch batch: The batch of events to extract the span links from.
    :rtype: list[~azure.core.tracing.Link]
    :return: A list of span links.
    """
    links = []
    try:
        for event in batch._internal_events:  # pylint: disable=protected-access
            # pylint: disable=protected-access
            msg = event if isinstance(event, AmqpAnnotatedMessage) else event._message
            link = get_span_link_from_message(msg)
            if link:
                links.append(link)
    except AttributeError:
        pass
    return links


def get_span_link_from_message(message: Union[AmqpAnnotatedMessage, Message]) -> Optional[Link]:
    """Create a span link from a message.

    This will extract the traceparent and tracestate from the message application properties and create span links
    based on these values.

    :param message: The message to extract the traceparent and tracestate from.
    :type message: ~azure.eventhub.amqp.AmqpAnnotatedMessage
     or ~azure.eventhub._pyamqp.message.Message or ~uamqp.Message
    :rtype: ~azure.core.tracing.Link
    :return: A span link.
    """
    headers = {}
    try:
        if message.application_properties:
            traceparent = (message.application_properties.get(TRACE_PARENT_PROPERTY, b"") or
                           message.application_properties.get(TRACE_DIAGNOSTIC_ID_PROPERTY, b""))
            if hasattr(traceparent, "decode"):
                traceparent = traceparent.decode(TRACE_PROPERTY_ENCODING)
            if traceparent:
                headers["traceparent"] = traceparent

            tracestate = message.application_properties.get(TRACE_STATE_PROPERTY, b"")
            if hasattr(tracestate, "decode"):
                tracestate = tracestate.decode(TRACE_PROPERTY_ENCODING)
            if tracestate:
                headers["tracestate"] = tracestate
    except AttributeError:
        return None
    return Link(headers)


def add_span_attributes(
        span: AbstractSpan,
        operation_type: TraceOperationTypes,
        client: Optional[ClientBase],
        message_count: int = 0
) -> None:
    """Add attributes to span based on the operation type.

    :param ~azure.core.tracing.AbstractSpan span: The span to add attributes to.
    :param TraceOperationTypes operation_type: The type of operation to add attributes for.
    :param ~azure.eventhub._client_base.ClientBase or None client: The client that is performing the operation.
    :param int message_count: The number of messages being processed.
    """

    span.add_attribute(TraceAttributes.TRACE_NAMESPACE_ATTRIBUTE, TraceAttributes.TRACE_NAMESPACE)
    span.add_attribute(TraceAttributes.TRACE_MESSAGING_SYSTEM_ATTRIBUTE, TraceAttributes.TRACE_MESSAGING_SYSTEM)
    span.add_attribute(TraceAttributes.TRACE_MESSAGING_OPERATION_ATTRIBUTE, operation_type)

    if message_count > 1:
        span.add_attribute(TraceAttributes.TRACE_MESSAGING_BATCH_COUNT_ATTRIBUTE, message_count)

    if operation_type in (TraceOperationTypes.PUBLISH, TraceOperationTypes.PROCESS):
        # Maintain legacy attributes for backwards compatibility.
        if client:
            span.add_attribute(TraceAttributes.LEGACY_TRACE_MESSAGE_BUS_DESTINATION_ATTRIBUTE, client._address.path)  # pylint: disable=protected-access
            span.add_attribute(TraceAttributes.LEGACY_TRACE_PEER_ADDRESS_ATTRIBUTE, client._address.hostname)  # pylint: disable=protected-access

    elif operation_type == TraceOperationTypes.RECEIVE:
        if client:
            span.add_attribute(TraceAttributes.TRACE_NET_PEER_NAME_ATTRIBUTE, client._address.hostname)  # pylint: disable=protected-access
            span.add_attribute(TraceAttributes.TRACE_MESSAGING_DESTINATION_ATTRIBUTE, client._address.path)  # pylint: disable=protected-access



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import os
import dotenv
import time
import logging
from logging.handlers import RotatingFileHandler
from concurrent.futures import ThreadPoolExecutor

from azure.eventhub import EventHubProducerClient, EventData

logger = logging.getLogger('SEND_PERF_TEST')
logger.setLevel(logging.INFO)
logger.addHandler(RotatingFileHandler("send_perf_test.log"))

dotenv.load_dotenv()
CONN_STRS = [
    os.environ["EVENT_HUB_CONN_STR_BASIC_NORTHEU"],
    os.environ["EVENT_HUB_CONN_STR_STANDARD_NORTHEU"],
    os.environ["EVENT_HUB_CONN_STR_BASIC_WESTUS2"],
    os.environ["EVENT_HUB_CONN_STR_STANDARD_WESTUS2"]
]
EVENTHUB_NAME = "pyamqp"

SINGLE_EVENT_SIZE_LIST = [512]
PARALLEL_THREAD_COUNT_LIST = [1]
FIXED_AMOUNT_OF_EVENTS = 100_000
RUN_DURATION = 30


def pre_prepare_client(client, data):
    client.create_batch()  # precall to retrieve sender link settings
    client.send_batch([EventData(data)])  # precall to set up the sender link


def send_batch_message(conn_str, eventhub_name, num_of_events, single_event_size, run_times=1, description=None):

    client = EventHubProducerClient.from_connection_string(
        conn_str=conn_str, eventhub_name=eventhub_name
    )

    data = b'a' * single_event_size
    perf_records = []
    pre_prepare_client(client, data)

    for _ in range(run_times):  # run run_times and calculate the avg performance
        start_time = time.time()
        batch = client.create_batch()
        for _ in range(num_of_events):
            try:
                batch.add(EventData(data))
            except ValueError:
                # Batch full
                client.send_batch(batch)
                batch = client.create_batch()
                batch.add(EventData(data))
        client.send_batch(batch)

        end_time = time.time()

        total_time = end_time - start_time
        speed = num_of_events / total_time
        perf_records.append(speed)

    client.close()
    avg_perf = round(sum(perf_records) / len(perf_records), 2)
    logger.info(
        "Method: {}, The average performance is {} events/s, throughput: {} bytes/s, run times: {}.\n"
        "Configs are: Num of events: {} events, Single message size: {} bytes.".format(
            description or "send_batch_message",
            avg_perf,
            avg_perf * single_event_size,
            run_times,
            num_of_events,
            single_event_size
        )
    )
    return avg_perf


def send_batch_message_worker_thread(client, data, run_flag):
    total_cnt = 0
    while run_flag[0]:
        batch = client.create_batch()
        try:
            while True:
                event_data = EventData(body=data)
                batch.add(event_data)
        except ValueError:
            client.send_batch(batch)
            total_cnt += len(batch)
    return total_cnt


def send_batch_message_in_parallel(conn_str, eventhub_name, single_event_size, parallel_thread_count=4, run_times=1, run_duration=30, description=None):

    perf_records = []

    for _ in range(run_times):

        futures = []
        clients = [
            EventHubProducerClient.from_connection_string(
                conn_str=conn_str, eventhub_name=eventhub_name
            ) for _ in range(parallel_thread_count)
        ]

        data = b'a' * single_event_size

        for client in clients:
            pre_prepare_client(client, data)

        with ThreadPoolExecutor(max_workers=parallel_thread_count) as executor:
            run_flag = [True]
            for i in range(parallel_thread_count):
                futures.append(
                    executor.submit(
                        send_batch_message_worker_thread,
                        clients[i],
                        data,
                        run_flag
                    )
                )

            time.sleep(run_duration)
            run_flag[0] = False
            perf_records.append(sum([future.result() for future in futures]) / run_duration)

            for client in clients:
                client.close()

    avg_perf = round(sum(perf_records) / len(perf_records), 2)

    logger.info(
        "Method: {}, The average performance is {} events/s, throughput: {} bytes/s, run times: {}.\n"
        "Configs are: Single message size: {} bytes, Parallel thread count: {} threads, Run duration: {} seconds.".format(
            description or "send_batch_message_in_parallel",
            avg_perf,
            avg_perf * single_event_size,
            run_times,
            single_event_size,
            parallel_thread_count,
            run_duration
        )
    )


if __name__ == '__main__':
    logger.info('------------------- START OF TEST -------------------')

    for conn_str in CONN_STRS:
        for single_event_size in SINGLE_EVENT_SIZE_LIST:
            print('-------------------  sending fixed amount of message  -------------------')
            send_batch_message(
                conn_str=conn_str,
                eventhub_name=EVENTHUB_NAME,
                num_of_events=FIXED_AMOUNT_OF_EVENTS,
                single_event_size=single_event_size,
                description='sending fixed amount message'
            )

        for parallel_thread_count in PARALLEL_THREAD_COUNT_LIST:
            for single_event_size in SINGLE_EVENT_SIZE_LIST:
                print('-------------------  multiple threads sending messages for a fixed period -------------------')
                send_batch_message_in_parallel(
                    conn_str=conn_str,
                    eventhub_name=EVENTHUB_NAME,
                    single_event_size=single_event_size,
                    parallel_thread_count=parallel_thread_count,
                    run_duration=RUN_DURATION,
                    description='multiple threads sending messages'
                )

    logger.info('------------------- END OF TEST -------------------')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import time
import asyncio
from collections import defaultdict

from ._test_base import _EventHubProcessorTest


class ProcessEventsBatchTest(_EventHubProcessorTest):
    def __init__(self, arguments):
        super().__init__(arguments)
        self._partition_event_count = defaultdict(int)

    def process_events_sync(self, partition_context, events):
        try:
            if events:
                if self.args.processing_delay:
                    delay_in_seconds = self.args.processing_delay / 1000
                    if self.args.processing_delay_strategy == 'sleep':
                        time.sleep(delay_in_seconds)
                    elif self.args.processing_delay_strategy == 'spin':
                        starttime = time.time()
                        while (time.time() - starttime) < delay_in_seconds:
                            pass
                
                # Consume properties and body.
                _ = [(list(e.body), str(e)) for e in events]

                if self.args.checkpoint_interval:
                    self._partition_event_count[partition_context.partition_id] += len(events)
                    if self._partition_event_count[partition_context.partition_id] >= self.args.checkpoint_interval:
                        partition_context.update_checkpoint()
                        self._partition_event_count[partition_context.partition_id] = 0
                for e in events:
                    self.event_raised_sync()
        except Exception as e:
            self.error_raised_sync(e)

    async def process_events_async(self, partition_context, events):
        try:
            if events:
                if self.args.processing_delay:
                    delay_in_seconds = self.args.processing_delay / 1000
                    if self.args.processing_delay_strategy == 'sleep':
                        await asyncio.sleep(delay_in_seconds)
                    elif self.args.processing_delay_strategy == 'spin':
                        starttime = time.time()
                        while (time.time() - starttime) < delay_in_seconds:
                            pass

                # Consume properties and body.
                _ = [(list(e.body), str(e)) for e in events]

                if self.args.checkpoint_interval:
                    self._partition_event_count[partition_context.partition_id] += len(events)
                    if self._partition_event_count[partition_context.partition_id] >= self.args.checkpoint_interval:
                        await partition_context.update_checkpoint()
                        self._partition_event_count[partition_context.partition_id] = 0
                for e in events:
                    await self.event_raised_async()
        except Exception as e:
            await self.error_raised_async(e)

    def process_error_sync(self, _, error):
        print(error)
        self.error_raised_sync(error)

    async def process_error_async(self, _, error):
        print(error)
        await self.error_raised_async(error)

    def start_events_sync(self) -> None:
        """
        Start the process for receiving events.
        """
        self.consumer.receive_batch(
            on_event_batch=self.process_events_sync,
            max_batch_size=self.args.max_batch_size,
            on_error=self.process_error_sync,
            max_wait_time=self.args.max_wait_time,
            prefetch=self.args.prefetch_count,
            starting_position="-1",  # "-1" is from the beginning of the partition.
        )

    async def start_events_async(self) -> None:
        """
        Start the process for receiving events.
        """
        await self.async_consumer.receive_batch(
            on_event_batch=self.process_events_async,
            max_batch_size=self.args.max_batch_size,
            on_error=self.process_error_async,
            max_wait_time=self.args.max_wait_time,
            prefetch=self.args.prefetch_count,
            starting_position="-1",  # "-1" is from the beginning of the partition.
        )

    @staticmethod
    def add_arguments(parser):
        super(ProcessEventsBatchTest, ProcessEventsBatchTest).add_arguments(parser)
        parser.add_argument('--max-batch-size', nargs='?', type=int, help='Maximum number of events to process in a single batch. Defaults to 100.', default=100)


#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

"""
An example to show receiving events from Event Hub partitions with custom starting position asynchronously.
"""
import os
import asyncio
from azure.eventhub.aio import EventHubConsumerClient, EventHubProducerClient
from azure.eventhub import EventData

CONNECTION_STR = os.environ["EVENT_HUB_CONN_STR"]
EVENTHUB_NAME = os.environ['EVENT_HUB_NAME']


async def on_partition_initialize(partition_context):
    # Put your code here.
    print("Partition: {} has been initialized.".format(partition_context.partition_id))


async def on_partition_close(partition_context, reason):
    # Put your code here.
    print("Partition: {} has been closed, reason for closing: {}.".format(
        partition_context.partition_id,
        reason
    ))


async def on_error(partition_context, error):
    # Put your code here. partition_context can be None in the on_error callback.
    if partition_context:
        print("An exception: {} occurred during receiving from Partition: {}.".format(
            partition_context.partition_id,
            error
        ))
    else:
        print("An exception: {} occurred during the load balance process.".format(error))


async def on_event(partition_context, event):
    # Put your code here.
    print("Received event: {} from partition: {}.".format(event.body_as_str(), partition_context.partition_id))
    await partition_context.update_checkpoint(event)


async def main():
    producer_client = EventHubProducerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        eventhub_name=EVENTHUB_NAME,
    )

    async with producer_client:
        event_data_batch_to_partition_0 = await producer_client.create_batch(partition_id='0')
        event_data_batch_to_partition_0.add(EventData("First event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Second event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Third event in partition 0"))
        event_data_batch_to_partition_0.add(EventData("Forth event in partition 0"))
        await producer_client.send_batch(event_data_batch_to_partition_0)

    consumer_client = EventHubConsumerClient.from_connection_string(
        conn_str=CONNECTION_STR,
        consumer_group='$Default',
        eventhub_name=EVENTHUB_NAME,
    )

    partition_0_prop = await consumer_client.get_partition_properties("0")
    partition_0_last_enqueued_sequence_number = partition_0_prop["last_enqueued_sequence_number"]

    # client will receive messages from the position of the third from last event on partition 0.
    starting_position = {
        "0": partition_0_last_enqueued_sequence_number - 3,
    }

    async with consumer_client:
        await consumer_client.receive(
            partition_id="0",
            on_event=on_event,
            on_partition_initialize=on_partition_initialize,
            on_partition_close=on_partition_close,
            on_error=on_error,
            starting_position=starting_position
        )


if __name__ == '__main__':
    asyncio.run(main())



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# pylint: disable=too-many-lines
from typing import Callable, cast, TYPE_CHECKING
from enum import Enum

from ._encode import encode_payload
from .utils import get_message_encoded_size
from .error import AMQPError
from .message import Header, Properties

if TYPE_CHECKING:
    from ..amqp._amqp_message import AmqpAnnotatedMessage

def _encode_property(value):
    try:
        return value.encode("UTF-8")
    except AttributeError:
        return value


class MessageState(Enum):
    WaitingToBeSent = 0
    WaitingForSendAck = 1
    SendComplete = 2
    SendFailed = 3
    ReceivedUnsettled = 4
    ReceivedSettled = 5

    def __eq__(self, __o: object) -> bool:
        try:
            return self.value == cast(Enum, __o).value
        except AttributeError:
            return super().__eq__(__o)


class MessageAlreadySettled(Exception):
    pass


DONE_STATES = (MessageState.SendComplete, MessageState.SendFailed)
RECEIVE_STATES = (MessageState.ReceivedSettled, MessageState.ReceivedUnsettled)
PENDING_STATES = (MessageState.WaitingForSendAck, MessageState.WaitingToBeSent)


class LegacyMessage(object):  # pylint: disable=too-many-instance-attributes
    def __init__(self, message, **kwargs):
        self._message: "AmqpAnnotatedMessage" = message
        self.state = MessageState.SendComplete
        self.idle_time = 0
        self.retries = 0
        self._settler = kwargs.get("settler")
        self._encoding = kwargs.get("encoding")
        self.delivery_no = kwargs.get("delivery_no")
        self.delivery_tag = kwargs.get("delivery_tag") or None
        self.on_send_complete = None
        self.properties = (
            LegacyMessageProperties(self._message.properties)
            if self._message.properties
            else None
        )
        self.application_properties = (
            self._message.application_properties
            if self._message.application_properties and any(self._message.application_properties)
            else None
        )
        self.annotations = (
            self._message.annotations
            if self._message.annotations and any(self._message.annotations)
            else None
        )
        self.header = (
            LegacyMessageHeader(self._message.header) if self._message.header else None
        )
        self.footer = self._message.footer
        self.delivery_annotations = self._message.delivery_annotations
        if self._settler:
            self.state = MessageState.ReceivedUnsettled
        elif self.delivery_no:
            self.state = MessageState.ReceivedSettled
        self._to_outgoing_amqp_message: Callable = kwargs.get(
            "to_outgoing_amqp_message"
        )

    def __str__(self):
        return str(self._message)

    def _can_settle_message(self):
        if self.state not in RECEIVE_STATES:
            raise TypeError("Only received messages can be settled.")
        if self.settled:
            return False
        return True

    @property
    def settled(self):
        if self.state == MessageState.ReceivedUnsettled:
            return False
        return True

    def get_message_encoded_size(self):
        return get_message_encoded_size(self._to_outgoing_amqp_message(self._message))

    def encode_message(self):
        output = bytearray()
        # to maintain the same behavior as uamqp, app prop values will not be decoded
        self.application_properties = self._message.application_properties.copy()
        encode_payload(output, self._to_outgoing_amqp_message(self._message))
        return bytes(output)

    def get_data(self):
        return self._message.body

    def gather(self):
        if self.state in RECEIVE_STATES:
            raise TypeError("Only new messages can be gathered.")
        if not self._message:
            raise ValueError("Message data already consumed.")
        if self.state in DONE_STATES:
            raise MessageAlreadySettled()
        return [self]

    def get_message(self):
        return self._to_outgoing_amqp_message(self._message)

    def accept(self):
        if self._can_settle_message():
            self._settler.settle_messages(self.delivery_no, "accepted")
            self.state = MessageState.ReceivedSettled
            return True
        return False

    def reject(self, condition=None, description=None, info=None):
        if self._can_settle_message():
            self._settler.settle_messages(
                self.delivery_no,
                "rejected",
                error=AMQPError(
                    condition=condition, description=description, info=info
                ),
            )
            self.state = MessageState.ReceivedSettled
            return True
        return False

    def release(self):
        if self._can_settle_message():
            self._settler.settle_messages(self.delivery_no, "released")
            self.state = MessageState.ReceivedSettled
            return True
        return False

    def modify(self, failed, deliverable, annotations=None):
        if self._can_settle_message():
            self._settler.settle_messages(
                self.delivery_no,
                "modified",
                delivery_failed=failed,
                undeliverable_here=deliverable,
                message_annotations=annotations,
            )
            self.state = MessageState.ReceivedSettled
            return True
        return False


class LegacyBatchMessage(LegacyMessage):
    batch_format = 0x80013700
    max_message_length = 1024 * 1024
    size_offset = 0


class LegacyMessageProperties(object):  # pylint: disable=too-many-instance-attributes
    def __init__(self, properties):
        self.message_id = _encode_property(properties.message_id)
        self.user_id = _encode_property(properties.user_id)
        self.to = _encode_property(properties.to)
        self.subject = _encode_property(properties.subject)
        self.reply_to = _encode_property(properties.reply_to)
        self.correlation_id = _encode_property(properties.correlation_id)
        self.content_type = _encode_property(properties.content_type)
        self.content_encoding = _encode_property(properties.content_encoding)
        self.absolute_expiry_time = properties.absolute_expiry_time
        self.creation_time = properties.creation_time
        self.group_id = _encode_property(properties.group_id)
        self.group_sequence = properties.group_sequence
        self.reply_to_group_id = _encode_property(properties.reply_to_group_id)

    def __str__(self):
        return str(
            {
                "message_id": self.message_id,
                "user_id": self.user_id,
                "to": self.to,
                "subject": self.subject,
                "reply_to": self.reply_to,
                "correlation_id": self.correlation_id,
                "content_type": self.content_type,
                "content_encoding": self.content_encoding,
                "absolute_expiry_time": self.absolute_expiry_time,
                "creation_time": self.creation_time,
                "group_id": self.group_id,
                "group_sequence": self.group_sequence,
                "reply_to_group_id": self.reply_to_group_id,
            }
        )

    def get_properties_obj(self):
        return Properties(
            self.message_id,
            self.user_id,
            self.to,
            self.subject,
            self.reply_to,
            self.correlation_id,
            self.content_type,
            self.content_encoding,
            self.absolute_expiry_time,
            self.creation_time,
            self.group_id,
            self.group_sequence,
            self.reply_to_group_id,
        )


class LegacyMessageHeader(object):
    def __init__(self, header):
        self.delivery_count = header.delivery_count or 0
        self.time_to_live = header.time_to_live
        self.first_acquirer = header.first_acquirer
        self.durable = header.durable
        self.priority = header.priority

    def __str__(self):
        return str(
            {
                "delivery_count": self.delivery_count,
                "time_to_live": self.time_to_live,
                "first_acquirer": self.first_acquirer,
                "durable": self.durable,
                "priority": self.priority,
            }
        )

    def get_header_obj(self):
        # TODO: uamqp returned object has property: `time_to_live`.
        # This Header has `ttl`.
        return Header(
            self.durable,
            self.priority,
            self.time_to_live,
            self.first_acquirer,
            self.delivery_count,
        )



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from ._test_base import _SendTest


class SendEventBatchTest(_SendTest):
    def __init__(self, arguments):
        super().__init__(arguments)

    def run_batch_sync(self):
        batch = self.producer.create_batch()
        for _ in range(self.args.batch_size):
            batch.add(self._build_event())
        self.producer.send_batch(batch)
        return self.args.batch_size

    async def run_batch_async(self):
        batch = await self.async_producer.create_batch()
        for _ in range(self.args.batch_size):
            batch.add(self._build_event())
        await self.async_producer.send_batch(batch)
        return self.args.batch_size
